"""
This module contains all the common logic of
the Suzieq plugins
"""

from importlib import import_module
from importlib.util import find_spec
from inspect import getmembers, getmro, isclass
from pkgutil import iter_modules
from typing import Dict, Type

from suzieq.shared.utils import get_sq_install_dir


class SqPlugin:
    """SqPlugin is the base common class inherited by all the
    Suzieq plugins
    """

    def __init__(self) -> None:
        pass

    @classmethod
    def get_plugins(cls,
                    plugin_name: str = None,
                    search_pkg: str = None) -> Dict[str, Type]:
        """Discover all the plugins in the search_pkg package, inheriting
        the current base class.

        Args:
            plugin_name (str): The name of a specific plugin to extract
            search_pkg (str): The package where to look for the plugins.
                Note that if the specified package contains other packages,
                the function will scan only their content, without accessing
                to other nested packages and ignoring all the files inside
                search_pkg.

        Returns:
            Dict[str, Type]: a dictionary containing all
                the descendants of the base class in the search_pkg,
                with format {'module name': PluginClass}
        """
        classes = {}

        if not search_pkg:
            search_pkg = '.'.join(cls.__module__.split('.')[:-1])

        # Collect the list of packages where to search
        sq_install_dir = '/'.join(get_sq_install_dir().split('/')[:-1])

        search_path = (f"{sq_install_dir}/{search_pkg.replace('.', '/')}")
        packages = [f'{search_pkg}.{m.name}'
                    for m in iter_modules([search_path])
                    if m.ispkg]
        if not packages:
            packages.append(search_pkg)
            use_pkg_name = False
        else:
            use_pkg_name = True

        for pkg in packages:
            found_plugin = False
            if use_pkg_name:
                use_name = pkg.split('.')[-1]
            pspec = find_spec(pkg)
            if pspec and pspec.loader:
                try:
                    mfound = [x.split('.')[0] for x in pspec.loader.contents()
                              if not x.startswith('_') and x.endswith(".py")]
                except Exception:  # pylint: disable=broad-except
                    mfound = []
                for minfo in mfound:
                    if not use_pkg_name:
                        use_name = minfo.split('.')[0]
                        mname = f'{pkg}.{use_name}'
                    else:
                        mname = f'{pkg}.{minfo}'
                    if plugin_name and plugin_name != use_name:
                        continue
                    mod = import_module(mname)
                    for mbr in getmembers(mod, isclass):
                        if (mbr[1].__module__ == mname
                           and getmro(mbr[1])[1] == cls
                           and issubclass(mbr[1], cls)
                           and mbr[1] != cls):
                            classes[use_name] = mbr[1]
                            found_plugin = True
            if plugin_name and found_plugin:
                break

        return classes



from nubia import command
from suzieq.cli.nubia_patch import argument

from suzieq.cli.sqcmds.command import SqTableCommand
from suzieq.sqobjects.sqPoller import SqPollerObj


@command("sqPoller", help="Act on SqPoller data")
@argument("service", description="Service(s), space separated")
@argument("status", description="status of service to match",
          choices=["all", "pass", "fail"])
@argument('poll_period_exceeded',
          description="filter if poll period exceeded",
          choices=['True', 'False'])
class SqPollerCmd(SqTableCommand):
    """Information about the poller"""

    def __init__(
            self,
            engine: str = "",
            hostname: str = "",
            start_time: str = "",
            end_time: str = "",
            view: str = "",
            namespace: str = "",
            format: str = "",  # pylint: disable=redefined-builtin
            query_str: str = "",
            columns: str = "default",
            service: str = '',
            status: str = '',
            poll_period_exceeded: str = '',
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=SqPollerObj,
        )

        poll_period_exceeded = str(poll_period_exceeded)
        if poll_period_exceeded == "False":
            poll_period_exceeded = "0"
        elif poll_period_exceeded:
            poll_period_exceeded = "!0"

        self.lvars = {
            'service': service.split(),
            'status': status,
            'pollExcdPeriodCount': poll_period_exceeded,
        }



import numpy as np
import pandas as pd

from suzieq.engines.pandas.engineobj import SqPandasEngine
from suzieq.shared.utils import build_query_str, humanize_timestamp


class BgpObj(SqPandasEngine):
    '''Backend class to handle manipulating BGP table with pandas'''

    @staticmethod
    def table_name():
        '''Table name'''
        return 'bgp'

    def get(self, **kwargs):
        """Replacing the original interface name in returned result"""

        columns = kwargs.pop('columns', ['default'])
        vrf = kwargs.pop('vrf', None)
        peer = kwargs.pop('peer', None)
        hostname = kwargs.pop('hostname', None)
        user_query = kwargs.pop('query_str', None)
        afi_safi = kwargs.pop('afiSafi', '')

        addnl_fields = ['origPeer']
        sch = self.schema
        fields = sch.get_display_fields(columns)

        self._add_active_to_fields(kwargs.get('view', self.iobj.view), fields,
                                   addnl_fields)
        if columns == ['*']:
            fields.remove('origPeer')

        for col in ['peerIP', 'updateSource', 'state', 'namespace', 'vrf',
                    'peer', 'hostname']:
            if col not in fields:
                addnl_fields.append(col)

        if afi_safi and afi_safi not in fields:
            addnl_fields.append('afiSafi')

        user_query_cols = self._get_user_query_cols(user_query)
        addnl_fields += [x for x in user_query_cols if x not in addnl_fields]

        try:
            df = super().get(addnl_fields=addnl_fields, columns=fields,
                             **kwargs)
        except KeyError as ex:
            if ('afi' in str(ex)) or ('safi' in str(ex)):
                df = pd.DataFrame(
                    {'error':
                     ['ERROR: Migrate BGP data first using sq-coalescer']})
                return df

        if df.empty:
            # We cannot add df[fields] here because we haven't added the
            # augmented columns yet and so can fail.
            return df

        if afi_safi or 'afiSafi' in fields or (columns == ['*']):
            df['afiSafi'] = df['afi'] + ' ' + df['safi']
        query_str = build_query_str([], sch, vrf=vrf, peer=peer,
                                    hostname=hostname, afiSafi=afi_safi,
                                    ignore_regex=False)
        if 'peer' in df.columns:
            df['peer'] = np.where(df['origPeer'] != "",
                                  df['origPeer'], df['peer'])

        if 'asndot' in fields:
            df['asndot'] = df.asn.apply(lambda x: f'{int(x/65536)}.{x%65536}')

        if 'peerAsndot' in fields:
            df['peerAsndot'] = df.peerAsn.apply(
                lambda x: f'{int(x/65536)}.{x%65536}')

        # Convert old data into new 2.0 data format
        if 'peerHostname' in df.columns:
            mdf = self._get_peer_matched_df(df)
        else:
            mdf = df

        mdf = self._handle_user_query_str(mdf, user_query)

        if query_str:
            return mdf.query(query_str).reset_index(drop=True)[fields]
        else:
            return mdf.reset_index(drop=True)[fields]

    def summarize(self, **kwargs) -> pd.DataFrame:
        """Summarize key information about BGP"""

        self._init_summarize(**kwargs)
        if self.summary_df.empty or ('error' in self.summary_df.columns):
            return self.summary_df

        self.summary_df['afiSafi'] = (
            self.summary_df['afi'] + ' ' + self.summary_df['safi'])

        afi_safi_count = self.summary_df.groupby(by=['namespace'])['afiSafi'] \
                                        .nunique()

        self.summary_df = self.summary_df \
                              .set_index(['namespace', 'hostname', 'vrf',
                                          'peer']) \
                              .query('~index.duplicated(keep="last")') \
                              .reset_index()
        self.ns = {i: {} for i in self.summary_df['namespace'].unique()}
        self.nsgrp = self.summary_df.groupby(by=["namespace"],
                                             observed=True)

        self._summarize_on_add_field = [
            ('deviceCnt', 'hostname', 'nunique'),
            ('totalPeerCnt', 'peer', 'count'),
            ('uniqueAsnCnt', 'asn', 'nunique'),
            ('uniqueVrfsCnt', 'vrf', 'nunique')
        ]

        self._summarize_on_add_with_query = [
            ('failedPeerCnt', 'state == "NotEstd"', 'peer'),
            ('iBGPPeerCnt', 'asn == peerAsn', 'peer'),
            ('eBGPPeerCnt', 'asn != peerAsn', 'peer'),
            ('rrClientPeerCnt', 'rrclient.str.lower() == "true"', 'peer',
             'count'),
        ]

        self._gen_summarize_data()

        # pylint: disable=expression-not-assigned
        {self.ns[i].update({'activeAfiSafiCnt': afi_safi_count[i]})
         for i in self.ns.keys()}
        self.summary_row_order.append('activeAfiSafiCnt')

        self.summary_df['estdTime'] = humanize_timestamp(
            self.summary_df.estdTime,
            self.cfg.get('analyzer', {}).get('timezone', None))

        self.summary_df['estdTime'] = (
            self.summary_df['timestamp'] - self.summary_df['estdTime'])
        self.summary_df['estdTime'] = self.summary_df['estdTime'] \
                                          .apply(lambda x: x.round('s'))
        # Now come the BGP specific ones
        established = self.summary_df.query("state == 'Established'") \
            .groupby(by=['namespace'])

        uptime = established["estdTime"]
        rx_updates = established["updatesRx"]
        tx_updates = established["updatesTx"]
        self._add_stats_to_summary(uptime, 'upTimeStat')
        self._add_stats_to_summary(rx_updates, 'updatesRxStat')
        self._add_stats_to_summary(tx_updates, 'updatesTxStat')

        self.summary_row_order.extend(['upTimeStat', 'updatesRxStat',
                                       'updatesTxStat'])

        self._post_summarize()
        return self.ns_df.convert_dtypes()

    def _get_peer_matched_df(self, df) -> pd.DataFrame:
        """Get a BGP dataframe that also contains a session's matching peer"""

        if 'peerHostname' not in df.columns:
            return df

        # We have to separate out the Established and non Established entries
        # for the merge. Otherwise we end up with a mess
        df_1 = df[['namespace', 'hostname', 'vrf', 'peer', 'peerIP',
                   'updateSource']] \
            .drop_duplicates() \
            .reset_index(drop=True)
        df_2 = df[['namespace', 'hostname', 'vrf', 'updateSource']] \
            .drop_duplicates() \
            .reset_index(drop=True)

        mdf = df_1.merge(df_2,
                         left_on=['namespace', 'peerIP'],
                         right_on=['namespace', 'updateSource'],
                         suffixes=('', '_y')) \
            .drop_duplicates(subset=['namespace', 'hostname', 'vrf',
                                     'peerIP']) \
            .rename(columns={'hostname_y': 'peerHost'}) \
            .fillna(value={'peerHostname': '', 'peerHost': ''}) \
            .reset_index(drop=True)

        df = df.merge(mdf[['namespace', 'hostname', 'vrf', 'peer',
                           'peerHost']],
                      on=['namespace', 'hostname', 'vrf', 'peer'], how='left')

        df['peerHostname'] = np.where((df['peerHostname'] == '') &
                                      (df['state'] == "Established"),
                                      df['peerHost'],
                                      df['peerHostname'])
        df = df.fillna(value={'peerHostname': ''}) \
            .drop(columns=['peerHost'])

        for i in df.select_dtypes(include='category'):
            df[i].cat.add_categories('', inplace=True)

        return df

    def aver(self, **kwargs) -> pd.DataFrame:
        """BGP Assert"""

        def _check_if_state(row, if_df):

            if not if_df.empty:
                thisif = if_df.query(f'namespace=="{row.namespace}" and '
                                     f'hostname=="{row.hostname}" and '
                                     f'ifname=="{row.ifname}"')
                if not thisif.empty:
                    if thisif.adminState.unique()[0] != 'up':
                        return ['interface admin down']
                    elif thisif.state.unique()[0] != 'up':
                        return ['interface down']
                    else:
                        return []

            return []

        assert_cols = ["namespace", "hostname", "vrf", "peer", "peerHostname",
                       "afi", "safi", "asn", "state", "peerAsn", "bfdStatus",
                       "reason", "notificnReason", "afisAdvOnly", 'ifname',
                       "afisRcvOnly", "peerIP", "updateSource", "timestamp"]

        kwargs.pop("columns", None)  # Loose whatever's passed
        result = kwargs.pop("result", 'all')
        state = kwargs.pop('state', '!dynamic')

        df = self.get(columns=assert_cols, state=state, **kwargs)
        if 'error' in df:
            return df

        if df.empty:
            if result != "pass":
                df['result'] = 'fail'
                df['assertReason'] = 'No data'
            return df

        df = self._get_peer_matched_df(df)
        if df.empty:
            if result != "pass":
                df['result'] = 'fail'
                df['assertReason'] = 'No data'
            return df

        # We can get rid of sessions with duplicate peer info since we're
        # interested only in session info here
        df = df.drop_duplicates(
            subset=['namespace', 'hostname', 'vrf', 'peer'])

        failed_df = df.query("state != 'Established'").reset_index(drop=True)
        passed_df = df.query("state == 'Established'").reset_index(drop=True)

        # Get the interface information
        if_df = self._get_table_sqobj('interfaces').get(
            namespace=failed_df.namespace.unique().tolist(),
            hostname=failed_df.hostname.unique().tolist(),
            ifname=failed_df.ifname.unique().tolist(),
            columns=['namespace', 'hostname', 'ifname', 'state', 'adminState']
        )

        failed_df['assertReason'] = [[] for _ in range(len(failed_df))]
        passed_df['assertReason'] = [[] for _ in range(len(passed_df))]

        if not failed_df.empty:
            # For not established entries, check if route/ARP entry exists
            failed_df['assertReason'] += failed_df.apply(
                _check_if_state, args=(if_df,), axis=1)

            failed_df['assertReason'] += failed_df.apply(
                lambda x: ["asn mismatch"]
                if (x['peerHostname'] and ((x["asn"] != x["peerAsn_y"]) or
                                           (x['asn_y'] != x['peerAsn'])))
                else [], axis=1)

            failed_df['assertReason'] += failed_df.apply(
                lambda x: [f"{x['reason']}:{x['notificnReason']}"]
                if ((x['reason'] and x['reason'] != 'None' and
                     x['reason'] != "No error"))
                else [], axis=1)

            failed_df['result'] = 'fail'

        # Get list of peer IP addresses for peer not in Established state
        # Returning to performing checks even if we didn't get LLDP/Intf info

        if not passed_df.empty:
            passed_df['assertReason'] += passed_df.apply(
                lambda x: ['Not all Afi/Safis enabled']
                if x['afisAdvOnly'].any() or x['afisRcvOnly'].any() else [],
                axis=1)

            passed_df['result'] = passed_df.apply(
                lambda x: 'pass'
                if len(x.assertReason) == 0 else 'fail',
                axis=1)

        df = pd.concat([failed_df, passed_df])

        result_df = df[['namespace', 'hostname', 'vrf', 'peer', 'asn',
                        'peerAsn', 'state', 'peerHostname', 'result',
                        'assertReason', 'timestamp']] \
            .explode(column="assertReason") \
            .fillna({'assertReason': '-'})

        if result == "fail":
            result_df = result_df.query('assertReason != "-"')
        elif result == "pass":
            result_df = result_df.query('assertReason == "-"')

        return result_df.reset_index(drop=True)



import time
from datetime import timedelta
from nubia import command

import pandas as pd

from suzieq.cli.nubia_patch import argument
from suzieq.cli.sqcmds.command import SqTableCommand
from suzieq.sqobjects.bgp import BgpObj


@command("bgp", help="Act on BGP data")
@argument("vrf", description="VRF(s), space separated")
@argument("state", description="status of the session to match",
          choices=["Established", "NotEstd", "dynamic"])
@argument("peer",
          description=("IP address(es), in quotes, or the interface name(s), "
                       "space separated"))
@argument("afiSafi", description="AFI SAFI string to filter by")
class BgpCmd(SqTableCommand):
    """BGP protocol information"""

    def __init__(
            self,
            engine: str = "",
            hostname: str = "",
            start_time: str = "",
            end_time: str = "",
            view: str = "",
            namespace: str = "",
            format: str = "",  # pylint: disable=redefined-builtin
            columns: str = "default",
            query_str: str = ' ',
            vrf: str = '',
            state: str = '',
            peer: str = '',
            afiSafi: str = '',
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=BgpObj,
        )
        self.lvars = {
            'vrf': vrf.split(),
            'state': state,
            'peer': peer.split(),
            'afiSafi': afiSafi,
        }

    def _clean_output(self, df) -> pd.DataFrame:
        """Make upTime look good"""
        if df.empty:
            return df

        if "estdTime" in df.columns:
            df['estdTime'] = df.estdTime.apply(
                lambda x: str(timedelta(milliseconds=int(x))))
        elif 'upTimeStat' in df.columns:
            df.loc['upTimeStat'] = df \
              .loc['upTimeStat'] \
              .map(lambda x: [str(timedelta(milliseconds=int(i))) for i in x])

        return df.dropna(how='any')

    @command("assert")
    @argument("result", description="Show only assert that matches this value",
              choices=["all", "fail", "pass"])
    def aver(self, result: str = "all") -> pd.DataFrame:
        """Assert BGP is functioning properly"""

        now = time.time()

        df = self._invoke_sqobj(self.sqobj.aver,
                                namespace=self.namespace,
                                hostname=self.hostname,
                                result=result,
                                **self.lvars,
                                )
        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)

        return self._assert_gen_output(df)



from typing import Type, Dict
import urllib
import requests
import urllib3

import pandas as pd

from suzieq.engines.base_engine import SqEngineObj


class SqRestEngine(SqEngineObj):
    '''The basic class implementing the REST engine'''

    def __init__(self, baseobj):
        self.ctxt = baseobj.ctxt
        self.iobj = baseobj

    @property
    def name(self):
        return 'rest'

    def table_name(self):
        '''table name, retrieved from sqobject

        Does it matter that the table name is not a static method as it is
        in pandas?
        '''
        return self.iobj.table

    @classmethod
    def get_plugins(cls,
                    plugin_name: str = None,
                    search_pkg: str = None) -> Dict[str, Type]:
        '''Redefining the base implementation

        We don't need to define specific table plugins for REST engine
        '''
        return {plugin_name: cls}

    def get(self, **kwargs) -> pd.DataFrame:
        '''The catch all get method for all tables'''
        return self._get_response('show', **kwargs)

    def summarize(self, **kwargs):
        return self._get_response('summarize', **kwargs)

    def unique(self, **kwargs):
        return self._get_response('unique', **kwargs)

    def aver(self, **kwargs):
        return self._get_response('assert', **kwargs)

    def top(self, **kwargs):
        return self._get_response('top', **kwargs)

    def find(self, **kwargs):
        return self._get_response('find', **kwargs)

    def lpm(self, **kwargs):
        return self._get_response('lpm', **kwargs)

    def _get_response(self, verb: str, **kwargs) -> pd.DataFrame:
        """The work horse engine implementing the basic REST API query

        Args:
            verb ([str]]): The verb for which we're executing the query
        """

        # Weed out the unspecified parameters
        cmd_params = {'start_time': self.iobj.start_time,
                      'end_time': self.iobj.end_time,
                      'view': kwargs.get('view', self.iobj.view),
                      'namespace': kwargs.get('namespace',
                                              self.iobj.namespace),
                      'hostname': kwargs.get('hostname', self.iobj.hostname)}
        kwargs.update(cmd_params)
        kwargs = {k: v for k, v in kwargs.items() if v}

        query_params = urllib.parse.urlencode(kwargs, doseq=True)

        if not self.ctxt.rest_api_key:
            raise ValueError("No REST API key specified")

        # Suppress warning since most folks will deploy this without
        # verifiable certificates
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        query_params = (
            f'{query_params}&access_token={self.ctxt.rest_api_key}'
        )
        table = self.table_name()
        if table in ['routes', 'macs', 'interfaces']:
            table = table[:-1]
        if table == "tables":
            # Screwup on our side in naming the API
            table = "table"
        url = (
            f'{self.ctxt.rest_transport}://{self.ctxt.rest_server_ip}'
            f':{self.ctxt.rest_server_port}'
            '/api/v2/'
            f'{table}/'
            f'{verb}?'
            f'{query_params}')

        response = requests.get(url, verify=False, )
        if response.status_code != 200:
            if response.text:
                msg = response.json().get("detail", str(response.status_code))
                # Strip out the "ID=" part of the message
                msg = msg.rsplit('id=', 1)[0].strip()
                return pd.DataFrame(
                    {'error': [f'ERROR: {msg}']})

            return pd.DataFrame({'error': [f'{response.status_code}']})

        df = pd.DataFrame(response.json())
        return df



import ast
import time
from nubia import command
from suzieq.cli.nubia_patch import argument

from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.network import NetworkObj


@command("network", help="Act on network-wide data")
class NetworkCmd(SqCommand):
    """Advanced commands across all the network"""

    def __init__(
            self,
            engine: str = "",
            hostname: str = "",
            start_time: str = "",
            end_time: str = "",
            view: str = "",
            namespace: str = "",
            format: str = "",  # pylint: disable=redefined-builtin
            query_str: str = " ",
            columns: str = "default",
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=NetworkObj,
        )

    @command("find", help="find where an IP/MAC address is attached")
    @argument("address", type=str,
              description="IP/MAC addresses, in quotes, space separated")
    @argument("vrf", type=str,
              description="Find within this VRF, used for IP addr")
    @argument("vlan", type=str,
              description="Find MAC within this VLAN")
    def find(self, address: str = '', vrf: str = '',
             vlan: str = ''):
        """Find the network attach point of a given IP or MAC address.
        """
        now = time.time()

        df = self._invoke_sqobj(self.sqobj.find,
                                namespace=self.namespace,
                                hostname=self.hostname,
                                address=address.split(),
                                vlan=vlan,
                                vrf=vrf,
                                query_str=self.query_str,
                                )

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)

        return self._gen_output(df)

    @command("show", help="Deprecated. Use 'namespace show' instead")
    @argument("model", description="Device model(s), space separated")
    @argument("os", description='Device NOS(es), space separated')
    @argument('vendor', description='Device vendor(s), space separated')
    @argument('version', description='Device NOS version(s), space separated')
    def show(self, os: str = "", vendor: str = "", model: str = "",
             version: str = "") -> int:
        """Deprecated. Use 'namespace show' instead
        """
        # backward compatibility
        self._print_depracation_command_warning('namespace', 'show')
        now = time.time()

        df = self._invoke_sqobj(self.sqobj.get,
                                namespace=self.namespace, os=os.split(),
                                vendor=vendor.split(), model=model.split(),
                                version=version, query_str=self.query_str,
                                hostname=self.hostname, ignore_warning=True)

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)

        return self._gen_output(df)

    @command("summarize", help="Deprecated. Use 'namespace summarize' instead")
    def summarize(self, **kwargs):
        """Deprecated. Use 'namespace summarize' instead"""
        # backward compatibility
        self._print_depracation_command_warning('namespace', 'summarize')
        now = time.time()

        summarize_df = self._invoke_sqobj(
            self.sqobj.summarize, namespace=self.namespace,
            hostname=self.hostname, query_str=self.query_str,
            ignore_warning=True,
            **self.lvars,
            **kwargs
        )
        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        return self._gen_output(summarize_df, json_orient='columns')

    @command("unique", help="Deprecated. Use 'namespace unique' instead")
    @argument("count", description="include count of times a value is seen",
              choices=['True'])
    def unique(self, count='', **kwargs):
        """Deprecated. Use 'namespace unique' instead"""
        # backward compatibility
        self._print_depracation_command_warning('namespace', 'unique')
        now = time.time()

        df = self._invoke_sqobj(self.sqobj.unique,
                                hostname=self.hostname,
                                namespace=self.namespace,
                                query_str=self.query_str,
                                count=count,
                                ignore_warning=True,
                                **self.lvars,
                                **kwargs,
                                )

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        if 'error' in df.columns:
            return self._gen_output(df)

        if df.empty:
            return df

        if not count:
            return self._gen_output(df.sort_values(by=[df.columns[0]]),
                                    dont_strip_cols=True)
        else:
            return self._gen_output(
                df.sort_values(by=['numRows', df.columns[0]]),
                dont_strip_cols=True)

    @ command("top", help="Deprecated. Use 'namespace top' instead")
    @ argument("count", description="number of rows to return")
    @ argument("what", description="numeric field to get top values for")
    @ argument("reverse", description="return bottom n values",
               choices=['True', 'False'])
    def top(self, count: int = 5, what: str = '', reverse: str = 'False',
            **kwargs) -> int:
        """Deprecated. Use 'namespace top' instead

        Args:
            count (int, optional): Number of entries to return. Defaults to 5
            what (str, optional): Field name to use for largest/smallest val
            reverse (bool, optional): Reverse and return n smallest

        Returns:
            int: 0 or error code
        """
        # backward compatibility
        self._print_depracation_command_warning('namespace', 'top')
        now = time.time()

        df = self._invoke_sqobj(self.sqobj.top,
                                hostname=self.hostname,
                                namespace=self.namespace,
                                query_str=self.query_str,
                                what=what, count=count,
                                reverse=ast.literal_eval(reverse),
                                ignore_warning=True,
                                **self.lvars,
                                **kwargs,
                                )

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        if 'error' in df.columns:
            return self._gen_output(df)

        if not df.empty:
            df = self.sqobj.humanize_fields(df)
            return self._gen_output(df.sort_values(
                by=[what], ascending=(reverse == "True")),
                dont_strip_cols=True, sort=False)
        else:
            return self._gen_output(df)

    @ command("help", help="show help for a command")
    @ argument("command", description="command to show help for",
               choices=['find', 'show', 'summarize', 'top', 'unique'])
    # pylint: disable=redefined-outer-name
    def help(self, command: str = ''):
        return super().help(command)



"""
This module contains all the logic needed to start and monitor the coalescer
"""
import asyncio
import errno
import logging
from pathlib import Path
from asyncio.subprocess import Process
from typing import Dict

from suzieq.poller.controller.utils.proc_utils import monitor_process
from suzieq.shared.exceptions import PollingError
from suzieq.shared.utils import get_sq_install_dir

logger = logging.getLogger(__name__)


class CoalescerLauncher:
    """CoalescerLauncher is the component in charge of start and monitor the
    running coalescer
    """

    def __init__(self,
                 config_file: str,
                 cfg: Dict,
                 coalescer_bin: str = None,
                 max_attempts: int = 3) -> None:
        """Initialize an instance of the CoalescerLauncher

        Args:
            config_file (str): the Suzieq configuration file to pass
                to the coalescer
            cfg (dict): the Suzieq config dictionary
            coalescer_bin (str, optional): optional path to coalescer binary.
                Defaults to None.
            max_attempts (int, optional): the number of attepts to perform
                for starting the coalescer. Defaults to 3
        """
        self.coalescer_process = None

        self.config_file = config_file
        self.cfg = cfg

        self.max_attempts = max_attempts
        if self.max_attempts <= 0:
            raise PollingError(
                'The number of attempts for starting the coalescer cannot be '
                '0 or less'
            )

        self.coalescer_bin = coalescer_bin
        if not coalescer_bin:
            sq_path = get_sq_install_dir()
            self.coalescer_bin = Path(f'{sq_path}/utilities/sq_coalescer.py')

    @property
    def coalescer_pid(self) -> int:
        """The coalescer_pid is the PID of the coalescer process

        Returns:
            int: the pid of the coalescer process
        """

        return self.coalescer_process.pid if self.coalescer_process else None

    async def start_and_monitor_coalescer(self):
        """Start and monitor the coalescer
        """
        try:
            await self._monitor_coalescer()
        except asyncio.CancelledError:
            pass
        finally:
            # If the coalescer is still runnning we always need to terminate
            # it before exiting
            if self.coalescer_process and \
               self.coalescer_process.returncode is None:
                self.coalescer_process.terminate()
                try:
                    logger.warning('Waiting coalescer termination...')
                    await asyncio.wait_for(self.coalescer_process.wait(), 10)
                except asyncio.TimeoutError:
                    self.coalescer_process.kill()

    async def _monitor_coalescer(self):
        """This function calls _start_coalescer() and check if the process
        dies
        """

        starting_attempts = 0

        # We would like to attempt the coalescer startup self.max_attempts
        # times
        while starting_attempts < self.max_attempts:
            if not self.coalescer_process:
                logger.info('Starting Coalescer')
            elif self.coalescer_process.returncode == errno.EBUSY:
                logger.warning('Trying to start coalescer')
            else:
                logger.warning('Restarting the coalescer')

            # Try to start the coalescer process
            self.coalescer_process = await self._start_coalescer()
            if not self.coalescer_process:
                raise PollingError('Unable to start the coalescer')

            # Check if we have something from the stdout we need to log
            await monitor_process(self.coalescer_process, 'COALESCER')

            # When the coalescer exists with EBUSY error, it means
            # there already is another instance of the coalescer running
            # we are fine with that, but we need to make sure it is always
            # running, so we wait some time before retrying launching again
            # the coalescer
            if self.coalescer_process.returncode == errno.EBUSY:
                logger.warning(
                    "There is a running coalescer instance, "
                    "let's try again later"
                )
                await asyncio.sleep(10*60)
            else:
                # The coalescer died for any other reason let's increase the
                # attempts counter
                starting_attempts += 1
        logger.error(
            f'Maximum number of attempts reached {self.max_attempts}/'
            f'{self.max_attempts}, the coalescer keep on crashing')

    async def _start_coalescer(self) -> Process:
        """Start the coalescer

        Returns:
            Process: the process object of the started coalescer.
        """
        coalescer_args = [self.coalescer_bin]
        if self.config_file:
            coalescer_args += ['-c', self.config_file]

        try:
            process = await asyncio.create_subprocess_exec(
                *coalescer_args, stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE)
        except Exception as ex:  # pylint: disable=broad-except
            logger.error(f'ABORTING. Unable to start coalescer: {ex}')
            process = None

        return process



import time
from nubia import command
from suzieq.cli.nubia_patch import argument

from suzieq.cli.sqcmds.command import SqTableCommand
from suzieq.sqobjects.evpnVni import EvpnvniObj


@command("evpnVni", help="Act on EVPN VNI data")
@argument("vni", description="VNI ID(s), space separated")
@argument("priVtepIp", description="Primary VTEP IP(s), space separated")
class EvpnVniCmd(SqTableCommand):
    """EVPN information such as VNI/VLAN mapping, VTEP IPs etc."""

    def __init__(
            self,
            engine: str = "",
            hostname: str = "",
            start_time: str = "",
            end_time: str = "",
            view: str = "",
            namespace: str = "",
            format: str = "",  # pylint: disable=redefined-builtin
            query_str: str = " ",
            columns: str = "default",
            vni: str = '',
            priVtepIp: str = '',
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=EvpnvniObj,
        )
        self.lvars = {
            'vni': vni,
            'priVtepIp': priVtepIp.split()
        }

    @command("assert")
    @argument("result", description="Show only assert that matches this value",
              choices=["all", "fail", "pass"])
    def aver(self, result: str = 'all'):
        """Assert VXLAN Forwarding is functioning properly"""

        now = time.time()

        df = self._invoke_sqobj(self.sqobj.aver,
                                hostname=self.hostname,
                                namespace=self.namespace,
                                result=result,
                                **self.lvars,
                                )
        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)

        return self._assert_gen_output(df)



from nubia import command
from suzieq.cli.nubia_patch import argument

from suzieq.cli.sqcmds.command import SqTableCommand
from suzieq.sqobjects.vlan import VlanObj


@command('vlan', help="Act on vlan data")
@argument("state", description="State of VLAN to query",
          choices=['active', 'suspended'])
@argument("vlan",
          description="VLAN(s), space separated, can use <, >, <=, >=, !")
@argument('vlanName',
          description="VLAN name(s), space separated")
class VlanCmd(SqTableCommand):
    """Information about VLANs including interfaces belonging to a VLAN"""

    # pylint: disable=redefined-builtin
    def __init__(
            self,
            engine: str = '',
            hostname: str = '',
            start_time: str = '',
            end_time: str = '',
            view: str = '',
            namespace: str = '',
            query_str: str = ' ',
            format: str = "",
            columns: str = 'default',
            vlan: str = '',
            vlanName: str = '',
            state: str = '',
    ) -> None:
        super().__init__(engine=engine, hostname=hostname,
                         start_time=start_time, end_time=end_time,
                         view=view, namespace=namespace, query_str=query_str,
                         format=format, columns=columns, sqobj=VlanObj)
        self.lvars = {
            'vlan': vlan.split(),
            'vlanName': vlanName.split(),
            'state': state
        }



###
# Defines the schema class and the associated methods for suzieq tables.
###

import os
import logging
import json
from typing import List, Dict, Optional
import pyarrow as pa


class Schema:
    '''Schema class holding schemas of all tables and providing ops on them'''

    def __init__(self, schema_dir):

        self._init_schemas(schema_dir)
        if not self._schema:
            raise ValueError("Unable to get schemas")

    def _init_schemas(self, schema_dir: str):
        """Returns the schema definition which is the fields of a table"""

        schemas = {}
        phy_tables = {}
        types = {}

        logger = logging.getLogger(__name__)
        if not (schema_dir and os.path.exists(schema_dir)):
            logger.error(
                "Schema directory %s does not exist", schema_dir)
            raise Exception(f"Schema directory {schema_dir} does not exist")

        for root, _, files in os.walk(schema_dir):
            for topic in files:
                if not topic.endswith(".avsc"):
                    continue
                with open(root + "/" + topic, "r") as f:
                    data = json.loads(f.read())
                    table = data["name"]
                    schemas[table] = data["fields"]
                    types[table] = data.get('recordType', 'record')
                    phy_tables[data["name"]] = data.get("physicalTable", table)
            break

        self._schema = schemas
        self._phy_tables = phy_tables
        self._types = types

    def tables(self) -> List[str]:
        '''Returns list of tables for which we have schemas'''
        return list(self._schema.keys())

    def fields_for_table(self, table: str) -> List[str]:
        '''Returns list of fields in given table'''
        return [f['name'] for f in self._schema[table]]

    def get_raw_schema(self, table: str) -> Dict:
        '''Raw schema for given table, JSON'''
        return self._schema[table]

    def field_for_table(self, table: str, field: str) -> Optional[str]:
        '''Returns info about field in table if present'''
        for f in self._schema[table]:
            if f['name'] == field:
                return f
        return None

    def type_for_table(self, table: str) -> str:
        '''Return table type: counter, record, derived etc.'''
        return self._types[table]

    def key_fields_for_table(self, table: str) -> List[str]:
        '''Return key fields for given table'''
        # return [f['name'] for f in self._schema[table]
        # if f.get('key', None) is not None]
        return self._sort_fields_for_table(table, 'key')

    def augmented_fields_for_table(self, table: str,
                                   fields: List[str],
                                   recurse: int) -> List[str]:
        '''Return all augmented fields in given list of fields (all if empty)
        If recurse is non-zero, continue to recurse till all fields are
        resolved
        '''
        all_aug_fields = [x['name'] for x in self._schema.get(table, [])
                          if 'depends' in x]
        if not fields:
            aug_fields = all_aug_fields
        else:
            aug_fields = [x for x in all_aug_fields if x in fields]

        if not (fields or recurse):
            return aug_fields

        all_parents = []
        new_aug_fields = []
        for f in aug_fields:
            all_parents += self.get_parent_fields(table, f)
        if all_parents:
            new_aug_fields = self.augmented_fields_for_table(
                table, all_parents, recurse-1)
        if not new_aug_fields:
            return aug_fields
        else:
            aug_fields.extend([x for x in new_aug_fields
                               if x not in aug_fields])
            return aug_fields

    def sorted_display_fields_for_table(self, table: str,
                                        getall: bool = False) -> List[str]:
        '''Returns sorted list of default display fields'''
        return self._sort_fields_for_table(table, 'display', getall)

    def _sort_fields_for_table(self, table: str, tag: str,
                               getall: bool = False) -> List[str]:
        '''Returns sorted list of fields in table with given tag'''
        fields = self.fields_for_table(table)
        field_weights = {}
        for f_name in fields:
            field = self.field_for_table(table, f_name)
            if field.get(tag, None) is not None:
                field_weights[f_name] = field.get(tag, 1000)
            elif getall:
                field_weights[f_name] = 1000
        return list(sorted(field_weights.keys(),
                           key=lambda x: field_weights[x]))

    def array_fields_for_table(self, table: str) -> List[str]:
        '''Returns list of fields which are lists in table'''
        fields = self.fields_for_table(table)
        arrays = []
        for f_name in fields:
            field = self.field_for_table(table, f_name)
            if (isinstance(field['type'], dict) and
                    field['type'].get('type', None) == 'array'):
                arrays.append(f_name)
        return arrays

    def get_phy_table_for_table(self, table: str) -> Optional[str]:
        """Return the name of the underlying physical table"""
        if self._phy_tables:
            return self._phy_tables.get(table, table)
        return None

    def get_partition_columns_for_table(self, table: str) -> List[str]:
        """Return the list of partition columns for table"""
        if self._phy_tables:
            return self._sort_fields_for_table(table, 'partition')
        return []

    def get_arrow_schema(self, table: str) -> pa.schema:
        """Convert internal AVRO schema into PyArrow schema"""

        avro_sch = self._schema.get(table, None)
        if not avro_sch:
            raise AttributeError(f"No schema found for {table}")

        arsc_fields = []

        map_type = {
            "string": pa.string(),
            "long": pa.int64(),
            "int": pa.int32(),
            "double": pa.float64(),
            "float": pa.float32(),
            "timestamp": pa.int64(),
            "timedelta64[s]": pa.float64(),
            "boolean": pa.bool_(),
            "array.string": pa.list_(pa.string()),
            "array.nexthopList": pa.list_(pa.struct([('nexthop', pa.string()),
                                                     ('oif', pa.string()),
                                                     ('weight', pa.int32())])),
            "array.long": pa.list_(pa.int64()),
            "array.float": pa.list_(pa.float32()),
            "array.double": pa.list_(pa.float64()),
        }

        for fld in avro_sch:
            if "depends" in fld:
                # These are augmented fields, not in arrow"
                continue
            if isinstance(fld["type"], dict):
                if fld["type"]["type"] == "array":
                    if fld["type"]["items"]["type"] == "record":
                        avtype: str = "array.{}".format(fld["name"])
                    else:
                        avtype: str = "array.{}".format(
                            fld["type"]["items"]["type"])
                else:
                    # We don't support map yet
                    raise AttributeError
            else:
                avtype: str = fld["type"]

            arsc_fields.append(pa.field(fld["name"], map_type[avtype]))

        return pa.schema(arsc_fields)

    def get_parent_fields(self, table: str, field: str) -> List[str]:
        '''Get list of fields this augmented field depends upon'''
        avro_sch = self._schema.get(table, None)
        if not avro_sch:
            raise AttributeError(f"No schema found for {table}")

        for fld in avro_sch:
            if fld['name'] == field:
                if "depends" in fld:
                    return fld['depends'].split()
                else:
                    return []
        return []


class SchemaForTable:
    '''Class supporting operations on the schema for a given table'''

    def __init__(self, table, schema: Schema = None, schema_dir=None):
        if schema:
            if isinstance(schema, Schema):
                self._all_schemas = schema
            else:
                raise ValueError("Passing non-Schema type for schema")
        else:
            self._all_schemas = Schema(schema_dir=schema_dir)
        self._table = table
        if table not in self._all_schemas.tables():
            raise ValueError(f"Unknown table {table}, no schema found for it")

    @ property
    def type(self):
        '''Type of table'''
        return self._all_schemas.type_for_table(self._table)

    @ property
    def version(self):
        '''DB version for table'''
        return self._all_schemas.field_for_table(self._table,
                                                 'sqvers')['default']

    @ property
    def fields(self):
        '''Returns list of fields for table'''
        return self._all_schemas.fields_for_table(self._table)

    @ property
    def array_fields(self):
        '''Return list of array fields in table'''
        return self._all_schemas.array_fields_for_table(self._table)

    def get_phy_table(self):
        '''Get the name of the physical table backing this table'''
        return self._all_schemas.get_phy_table_for_table(self._table)

    def get_partition_columns(self):
        '''Get the Parquet partitioning columns'''
        return self._all_schemas.get_partition_columns_for_table(self._table)

    def key_fields(self) -> List[str]:
        '''Returns list of key fields for table'''
        return self._all_schemas.key_fields_for_table(self._table)

    def get_augmented_fields(self, fields) -> List[str]:
        '''Returns list of augmented fields, recursively resolving'''
        MAX_AUGMENTED_RECURSE = 2

        return self._all_schemas.augmented_fields_for_table(
            self._table, fields, MAX_AUGMENTED_RECURSE)

    def sorted_display_fields(self, getall: bool = False) -> List[str]:
        '''Returns sorted list of default display fields'''
        return self._all_schemas.sorted_display_fields_for_table(self._table,
                                                                 getall)

    def field(self, field):
        '''Returns info about the specified field in table'''
        return self._all_schemas.field_for_table(self._table, field)

    def get_display_fields(self, columns: list) -> list:
        """Return the list of display fields for the given table"""
        if columns == ["default"]:
            fields = self.sorted_display_fields()

            if "namespace" not in fields:
                fields.insert(0, "namespace")
        elif columns == ["*"]:
            fields = self.sorted_display_fields(getall=True)
            fields.remove('sqvers')
        else:
            wrong_fields = [f for f in columns if f not in self.fields]
            if wrong_fields:
                raise ValueError(
                    f'Unknown fields {wrong_fields} in {self._table}')
            fields = columns.copy()

        return fields

    def get_phy_table_for_table(self) -> str:
        """Return the underlying physical table for this logical table"""
        return self._all_schemas.get_phy_table_for_table(self._table)

    def get_raw_schema(self) -> Dict:
        '''Return the raw schema of table'''
        return self._all_schemas.get_raw_schema(self._table)

    def get_arrow_schema(self) -> pa.Schema:
        '''Get Pyarrow schema of table'''
        return self._all_schemas.get_arrow_schema(self._table)

    def get_parent_fields(self, field) -> List[str]:
        '''Get dependent fields for a given augmented field in table'''
        return self._all_schemas.get_parent_fields(self._table, field)



from nubia import command
from suzieq.cli.nubia_patch import argument

from suzieq.cli.sqcmds.command import SqTableCommand
from suzieq.sqobjects.arpnd import ArpndObj


@command("arpnd", help="Act on ARP/ND data")
@argument("prefix",
          description=("Show all the addresses in this "
                       "subnet prefix (in quotes)"))
@argument("address",
          description="IP Address(es), in quotes, space separated")
@argument("macaddr",
          description="MAC address(es), in quotes, space separated")
@argument("oif", description="Outgoing interface(s), space separated")
class ArpndCmd(SqTableCommand):
    """ARP/Neighbor Discovery information"""

    def __init__(
            self,
            engine: str = "",
            hostname: str = "",
            start_time: str = "",
            end_time: str = "",
            view: str = "",
            namespace: str = "",
            format: str = "",  # pylint: disable=redefined-builtin
            columns: str = "default",
            query_str: str = ' ',
            prefix: str = '',
            address: str = '',
            macaddr: str = '',
            oif: str = '',
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=ArpndObj,
        )
        self.lvars = {
            'prefix': prefix.split(),
            'ipAddress': address.split(),
            'macaddr': macaddr.split(),
            'oif': oif.split(),
        }



import pandas as pd
from pandas.core.dtypes.dtypes import DatetimeTZDtype

from suzieq.sqobjects.basicobj import SqObject
from suzieq.shared.utils import humanize_timestamp


class InterfacesObj(SqObject):
    '''The object providing access to the interfaces table'''

    def __init__(self, **kwargs):
        super().__init__(table='interfaces', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'ifname', 'columns',
                                'state', 'type', 'mtu', 'master', 'ifindex',
                                'vrf', 'portmode', 'vlan', 'query_str']
        self._valid_assert_args = self._valid_get_args + \
            ['what', 'matchval', 'result', 'ignore_missing_peer']
        self._valid_arg_vals = {
            'state': ['up', 'down', 'notConnected', '!up', '!down',
                      '!notConnected', ''],
            'result': ['all', 'pass', 'fail'],
        }
        self._unique_def_column = ['type']

    def humanize_fields(self, df: pd.DataFrame, _=None) -> pd.DataFrame:
        '''Humanize the timestamp and boot time fields'''
        if df.empty:
            return df

        if 'statusChangeTimestamp' in df.columns:
            if not isinstance(df.statusChangeTimestamp.dtype, DatetimeTZDtype):
                df['statusChangeTimestamp'] = humanize_timestamp(
                    df.statusChangeTimestamp,
                    self.cfg.get('analyzer', {}).get('timezone', None))

        return super().humanize_fields(df)



from ipaddress import ip_interface

from suzieq.sqobjects.basicobj import SqObject
from suzieq.shared.utils import (convert_macaddr_format_to_colon,
                                 validate_macaddr)
from suzieq.shared.utils import validate_network


class ArpndObj(SqObject):
    '''The object providing access to the arp/nd table'''

    def __init__(self, **kwargs):
        super().__init__(table='arpnd', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'ipAddress', 'prefix',
                                'columns', 'oif', 'macaddr', 'query_str']
        self._convert_args = {
            'macaddr': convert_macaddr_format_to_colon
        }

    def validate_get_input(self, **kwargs):
        if kwargs.get('prefix', []) and kwargs.get('ipAddress', []):
            raise AttributeError("Cannot specify address and prefix together")

        if kwargs.get('prefix', []):
            for p in kwargs['prefix']:
                if not validate_network(p):
                    raise ValueError("Invalid prefix specified")

        if kwargs.get('ipAddress', []):
            for a in kwargs['ipAddress']:
                try:
                    ip_interface(a)
                except ValueError:
                    raise ValueError("Invalid IP address specified")

        if kwargs.get('macaddr', []):
            for m in kwargs['macaddr']:
                if not validate_macaddr(m):
                    raise ValueError("Invalid mac address specified")

        super().validate_get_input(**kwargs)
        self._unique_def_column = ['ipAddress']



from ipaddress import ip_interface

import numpy as np
import pandas as pd

from suzieq.engines.pandas.engineobj import SqPandasEngine
from suzieq.shared.utils import (convert_macaddr_format_to_colon,
                                 build_query_str)


class AddressObj(SqPandasEngine):
    '''Backend to process/analyze Address table data'''

    @staticmethod
    def table_name():
        '''Table name'''
        return 'address'

    def addr_type(self, addr: list) -> list:
        '''Return IP version of address, or 0 if MAC addr'''
        rslt = []
        for a in addr:
            try:
                ipa = ip_interface(a)
                rslt.append(ipa._version)
            except ValueError:
                rslt.append(0)
        return rslt

    # pylint: disable=too-many-statements
    def get(self, **kwargs) -> pd.DataFrame:
        """Retrieve the dataframe that matches a given IPv4/v6/MAC address"""

        addr = kwargs.pop("address", [])
        prefix = kwargs.pop("prefix", [])
        columns = kwargs.pop("columns", [])
        ipvers = kwargs.pop("ipvers", "")
        user_query = kwargs.pop("query_str", "")

        if user_query:
            if user_query.startswith('"') and user_query.endswith('"'):
                user_query = user_query[1:-1]

        vrf = kwargs.pop("vrf", "")

        addnl_fields = []
        fields = self.schema.get_display_fields(columns)
        self._add_active_to_fields(kwargs.get('view', self.iobj.view),
                                   fields, addnl_fields)

        if prefix:
            addr_types = self.addr_type(prefix)
        else:
            addr_types = self.addr_type(addr)

        # Always include ip or mac addresses in the dataframe
        # if there is a filter on them

        for x in ['ipAddressList', 'ip6AddressList', 'macaddr']:
            if x not in fields:
                addnl_fields.append(x)

        user_query_cols = self._get_user_query_cols(user_query)
        addnl_fields += [x for x in user_query_cols if x not in addnl_fields]

        df = super().get(addnl_fields=addnl_fields, columns=fields,
                         **kwargs)

        if df.empty:
            return df

        if 'master' in df.columns:
            df = df.rename({'master': 'vrf'}, axis=1) \
                   .replace({'vrf': {'': 'default'}})
            df.loc[(df.vrf == 'bridge') |
                   ((df.ipAddressList.str.len() == 0)
                    & (df.ip6AddressList.str.len() == 0)),
                   'vrf'] = ''

        query_str = build_query_str([], self.schema, vrf=vrf)

        addrcols = []
        if 4 in addr_types or ipvers in ["v4", ""]:
            # df = df.explode('ipAddressList').fillna({'ipAddressList': ''})
            addrcols.append('ipAddressList')

        if 6 in addr_types or ipvers in ["v6", ""]:
            # df = df.explode('ip6AddressList').fillna({'ip6AddressList': ''})
            addrcols.append('ip6AddressList')

        if ('ipAddress' in columns or (columns == ['*'])) and not ipvers:
            ndf = pd.DataFrame(df[addrcols].agg(
                self._merge_address_cols, axis=1),
                columns=['ipAddress'])
            df = pd.concat([df, ndf], axis=1)

        v4addr = []
        v6addr = []
        filter_prefix = ''

        # Address and prefix filtering are mutual exclusive
        if addr:
            macaddr = []
            for i, a in enumerate(addr):
                if addr_types[i] == 0:
                    # convert the macaddr format to internal format
                    a = convert_macaddr_format_to_colon(a)
                    macaddr.append(a)
                elif addr_types[i] == 4:
                    if '/' not in a:
                        a += '/'
                    v4addr.append(a)
                elif addr_types[i] == 6:
                    if '/' not in a:
                        a += '/'
                    v6addr.append(a)

            # IMPORTANT: Don't mess with this order of query.
            # Some bug in pandas prevents it from working if
            # macaddr isn't first and your query
            # contains both a macaddr and an IP address.
            dfmac = dfv4 = dfv6 = pd.DataFrame()

            if macaddr:
                dfmac = df[df.macaddr.isin(macaddr)]

            if v4addr:
                dfv4 = df[df.ipAddressList.apply(
                    lambda x, addrs: any(a.startswith(tuple(addrs))
                                         for a in x), args=(v4addr,))]
            if v6addr:
                dfv6 = df[df.ip6AddressList.apply(
                    lambda x, addrs: any(a.startswith(tuple(addrs))
                                         for a in x), args=(v6addr,))]
            if v4addr or v6addr or macaddr:
                df = pd.concat([dfv4, dfv6, dfmac])
        elif prefix:
            for i, a in enumerate(prefix):
                if addr_types[i] == 4:
                    v4addr.append(a)
                elif addr_types[i] == 6:
                    v6addr.append(a)

            if v4addr:
                for a in v4addr:
                    query_str += (f'{filter_prefix} '
                                  f'@self._is_in_subnet(ipAddressList,"{a}")')
                    filter_prefix = 'or'
            if v6addr:
                for a in v6addr:
                    query_str += (f'{filter_prefix} '
                                  f'@self._is_in_subnet(ip6AddressList,"{a}")')
                    filter_prefix = 'or'

        if not query_str:
            if ipvers == "v4":
                query_str = 'ipAddressList.str.len() != 0'
            elif ipvers == "v6":
                query_str = 'ip6AddressList.str.len() != 0'
            elif ipvers == "l2":
                query_str = 'macaddr.str.len() != 0'

        if query_str:
            df = df.query(query_str)

        df = self._handle_user_query_str(df, user_query)
        return df.reset_index(drop=True)[fields]

    def unique(self, **kwargs) -> pd.DataFrame:
        """Specific here only if columns=vrf to filter out non-routed if"""
        columns = kwargs.pop('columns', [])
        ipvers = kwargs.pop('ipvers', '')

        if 'ipAddress' in columns:
            if 'v4' in ipvers:
                columns = ['ipAddressList']
            elif 'v6' in ipvers:
                columns = ['ip6AddressList']

        if columns == ['vrf']:
            vrf_query_str = ('(ipAddressList.str.len() != 0 or '
                             'ip6AddressList.str.len() != 0)')
            kwargs['get_query_str'] = vrf_query_str
        return super().unique(ipvers=ipvers, columns=columns, **kwargs)

    def summarize(self, **kwargs) -> pd.DataFrame:
        """Summarize address related info"""

        self._init_summarize(**kwargs)
        if self.summary_df.empty:
            return self.summary_df

        self._add_field_to_summary('hostname', 'nunique', 'deviceCnt')
        self._add_field_to_summary('hostname', 'count', 'addressCnt')
        self._add_field_to_summary('macaddr', 'nunique', 'uniqueIfMacCnt')

        v6df = self.summary_df.explode('ip6AddressList') \
            .dropna(how='any') \
            .query('~ip6AddressList.str.startswith("fe80")')
        if not v6df.empty:
            v6device = v6df.groupby(by=['namespace'])['hostname'].nunique()
            v6addr = v6df.groupby(by=['namespace'])['ip6AddressList'].nunique()
            for i in self.ns.keys():
                self.ns[i].update(
                    {'deviceWithv6AddressCnt': v6device.get(i, 0)})
                self.ns[i].update({'uniqueV6AddressCnt': v6addr.get(i, 0)})
        else:
            for i in self.ns.keys():
                self.ns[i].update({'deviceWithv6AddressCnt': 0})
                self.ns[i].update({'uniqueV6AddressCnt': 0})

        v4df = self.summary_df.explode('ipAddressList') \
            .dropna(how='any') \
            .query('ipAddressList.str.len() != 0')
        if not v4df.empty:
            v4device = v4df.groupby(by=['namespace'])['hostname'].nunique()
            v4addr = v4df.groupby(by=['namespace'])['ipAddressList'].nunique()
            for i in self.ns.keys():
                self.ns[i].update(
                    {'deviceWithv4AddressCnt': v4device.get(i, 0)})
                self.ns[i].update({'uniqueV4AddressCnt': v4addr.get(i, 0)})

            v4df['prefixlen'] = v4df.ipAddressList.str.split('/').str[1]

            # this doesn't work if we've filtered by namespace
            #  pandas complains about an index problem
            #  so instead we have the more complicated expression below
            #  they are equivalent
            # v4pfx = v4df.groupby(by=['namespace'])['prefixlen'] \
            #             .value_counts().rename('count').reset_index()
            v4pfx = v4df.groupby(by=['namespace', 'prefixlen'],
                                 as_index=False)['ipAddressList'].count() \
                .dropna()
            v4pfx = v4pfx.rename(columns={'ipAddressList': 'count'})
            v4pfx['count'] = v4pfx['count'].astype(int)
            v4pfx = v4pfx.sort_values(by=['count'], ascending=False)

            for i in self.ns.keys():
                cnts = []
                v4pfx[v4pfx['namespace'] == i].apply(
                    lambda x: cnts.append(x['prefixlen']), axis=1, args=cnts)
                self.ns[i].update({'subnetsUsed': cnts})
                cnts = []
                v4pfx[v4pfx['namespace'] == i].apply(
                    lambda x: cnts.append({x['prefixlen']: x['count']}),
                    axis=1, args=cnts)
                self.ns[i].update({'subnetTopCounts': cnts[:3]})
        else:
            for i in self.ns.keys():
                self.ns[i].update({'deviceWithv4AddressCnt': 0})
                self.ns[i].update({'uniqueV4AddressCnt': 0})
                self.ns[i].update({'subnetsUsed': []})
                self.ns[i].update({'subnetTopCounts': []})

        self.summary_row_order = ['deviceCnt', 'addressCnt',
                                  'uniqueV4AddressCnt', 'uniqueV6AddressCnt',
                                  'uniqueIfMacCnt',
                                  'deviceWithv4AddressCnt',
                                  'deviceWithv6AddressCnt', 'subnetsUsed',
                                  'subnetTopCounts']
        self._post_summarize(check_empty_col='addressCnt')
        return self.ns_df.convert_dtypes()

    def _merge_address_cols(self, df: pd.DataFrame) -> np.array:
        res = np.append(np.array([]), df.values[0])
        return np.append(res, df.values[1])



from nubia import command
from suzieq.cli.nubia_patch import argument

from suzieq.cli.sqcmds.command import SqTableCommand
from suzieq.sqobjects.topology import TopologyObj


@command("topology", help="build and act on topology data")
@argument("polled", description="Is the device polled by Suzieq",
          choices=['True', 'False'])
@argument("ifname", description="Interface name(s), space separated")
@argument("via",
          description="Protocol(s) via which nodes are connected, "
          "space separated")
@argument("vrf", description="VRF(s), space separated")
@argument("asn", description="BGP ASN(s), space separated")
@argument("area", description="OSPF Area(s), space separated")
@argument("afiSafi", description="BGP AFI SAFI lens to filter the topology")
@ argument("peerHostname",
           description="Peer hostname(s), space separated, "
           "space separated")
class TopologyCmd(SqTableCommand):
    """Information about the topology constructed from various protocols"""

    def __init__(
            self,
            engine: str = "",
            hostname: str = "",
            start_time: str = "",
            end_time: str = "",
            view: str = "",
            namespace: str = "",
            format: str = "",  # pylint: disable=redefined-builtin
            query_str: str = ' ',
            columns: str = "default",
            polled: str = '',
            ifname: str = '',
            via: str = '',
            vrf: str = '',
            asn: str = '',
            area: str = '',
            peerHostname: str = '',
            afiSafi: str = ''
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=TopologyObj
        )
        self.lvars = {
            'polled': str(polled),
            'ifname': ifname.split(),
            'asn': asn.split(),
            'area': area.split(),
            'via': via.split(),
            'vrf': vrf.split(),
            'peerHostname': peerHostname.split(),
            'afiSafi': afiSafi,
        }



from suzieq.sqobjects.basicobj import SqObject


class OspfNbrObj(SqObject):
    '''The object providing access to the ospf neighbor table

    For use from tables obj only. Use ospf object instead.
    '''

    def __init__(self, **kwargs):
        super().__init__(table='ospfNbr', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'columns', 'ifname',
                                'vrf', 'area', 'query_str']



from suzieq.engines.pandas.engineobj import SqPandasEngine


class SqpollerObj(SqPandasEngine):
    '''Backend class to handle manipulating sqpoller table with pandas'''

    @staticmethod
    def table_name():
        '''Table name'''
        return 'sqPoller'

    def get(self, **kwargs):
        '''Retrieve sqpoller data for the given filters'''

        status = kwargs.pop('status', '')
        poll_period_exceeded = kwargs.pop('pollExcdPeriodCount', '')
        columns = kwargs.pop('columns', [])
        addnl_fields = kwargs.pop('addnl_fields', [])

        fields = self.schema.get_display_fields(columns)
        if status and 'status' not in fields:
            addnl_fields.append('status')

        self._add_active_to_fields(kwargs.get('view', self.iobj.view), fields,
                                   addnl_fields)

        if poll_period_exceeded and 'pollExcdPeriodCount' not in fields:
            addnl_fields.append('pollExcdPeriodCount')

        df = super().get(columns=fields, **kwargs)

        # Additionally filter by status or exceeded polling period. We need to
        # filter after getting the data because otherwise the parquet filter
        # will pick the last entry which was not 0, for example, even if it
        # is not the latest, which is incorrect. Just like status=down
        add_filter = ''
        if status == "pass":
            add_filter = 'status == 0 or status == 200'
        elif status == "fail":
            add_filter = 'status != 0 and status != 200'

        if add_filter:
            add_prefix = ' and '
        else:
            add_prefix = ''

        if poll_period_exceeded:
            if poll_period_exceeded.startswith('!'):
                add_filter += f'{add_prefix}pollExcdPeriodCount != 0'
            else:
                add_filter += f'{add_prefix}pollExcdPeriodCount == 0'

        if not df.empty and add_filter:
            df = df.query(add_filter).reset_index(drop=True)

        return df.reset_index(drop=True)[fields]

    def summarize(self, **kwargs):
        '''Summarize poller operational state data'''
        self._summarize_on_add_field = [
            ('deviceCnt', 'hostname', 'nunique'),
            ('entriesCnt', 'hostname', 'count')
        ]

        self._summarize_on_add_list_or_count = [
            ('service', 'service'),
            ('status', 'status')
        ]

        self._summarize_on_add_stat = [
            ('pollExcdPeriodStat', '', 'pollExcdPeriodCount'),
        ]

        return super().summarize(**kwargs)



import pandas as pd

from suzieq.sqobjects.basicobj import SqObject
from suzieq.shared.utils import humanize_timestamp


class NamespaceObj(SqObject):
    '''The object providing access to the virtual table: namespace'''

    def __init__(self, **kwargs):
        super().__init__(table='namespace', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'version', 'os',
                                'model', 'vendor', 'columns', 'query_str']
        self._unique_def_column = ['namespace']

    def humanize_fields(self, df: pd.DataFrame, _=None) -> pd.DataFrame:
        '''Humanize the timestamp fields'''
        if df.empty:
            return df

        if 'lastUpdate' in df.columns:
            df['lastUpdate'] = humanize_timestamp(df.lastUpdate,
                                                  self.cfg.get('analyzer', {})
                                                  .get('timezone', None))

        return super().humanize_fields(df)



"""This module contains the base class for a Chunker.

Classes:
    Chunker: The duty of a chunker is to get the global inventory and split it
             in smaller chunks
"""
from abc import abstractmethod
from suzieq.poller.controller.base_controller_plugin import ControllerPlugin


class Chunker(ControllerPlugin):
    """Abstract class for a Chunker
    """

    @abstractmethod
    def chunk(self, glob_inv, n_chunks, **addl_params):
        """Split the global inventory in <n_chunks> chunks

        Args:
            glob_inv ([type]): global inventory to split
            n_chunks ([type]): number of chunks
            addl_parameters ([type]): custom parameters that each Chunker
                                      plugin can define
        """

    @classmethod
    def default_type(cls) -> str:
        return 'static'

    @classmethod
    def get_data_model(cls):
        """This is only temporary. In future release I will add chunker
        validation via pydantic
        """
        raise NotImplementedError



"""Netbox module

This module contains the methods to connect with a Netbox REST server
and retrieve the devices inventory

Classes:
    Netbox: this class dinamically retrieve the inventory from Netbox
"""
# pylint: disable=no-name-in-module
# pylint: disable=no-self-argument

import asyncio
import logging
from typing import Dict, List, Optional, Tuple, Union
from urllib.parse import urlparse

from pydantic import BaseModel, validator, Field

import aiohttp
from suzieq.poller.controller.inventory_async_plugin import \
    InventoryAsyncPlugin
from suzieq.poller.controller.source.base_source import Source, SourceModel
from suzieq.shared.utils import get_sensitive_data
from suzieq.shared.exceptions import InventorySourceError, SensitiveLoadError

_DEFAULT_PORTS = {'http': 80, 'https': 443}

logger = logging.getLogger(__name__)


class NetboxServerModel(BaseModel):
    """Model containing data to connect with Netbox server
    """
    host: str
    protocol: str
    port: str

    class Config:
        """pydantic configuration
        """
        extra = 'forbid'


class NetboxSourceModel(SourceModel):
    """Netbox source validation model
    """
    tag: Optional[str] = Field(default='suzieq')
    period: Optional[int] = Field(default=3600)
    token: str
    ssl_verify: Optional[bool] = Field(alias='ssl-verify')
    server: Union[str, NetboxServerModel] = Field(alias='url')
    run_once: Optional[bool] = Field(default=False, alias='run_once')

    @validator('server', pre=True)
    def validate_and_set(cls, url, values):
        """Validate the field 'url' and set the correct parameters
        """
        if isinstance(url, str):
            url_data = urlparse(url)
            host = url_data.hostname
            if not host:
                raise ValueError(f'Unable to parse hostname {url}')
            protocol = url_data.scheme or 'http'
            if protocol not in _DEFAULT_PORTS:
                raise ValueError(f'Unknown protocol {protocol}')
            port = url_data.port or _DEFAULT_PORTS.get(protocol)
            if not port:
                raise ValueError(f'Unable to parse port {url}')
            server = NetboxServerModel(host=host, port=port, protocol=protocol)
            ssl_verify = values['ssl_verify']
            if ssl_verify is None:
                if server.protocol == 'http':
                    ssl_verify = False
                else:
                    ssl_verify = True
            else:
                if server.protocol == 'http' and ssl_verify:
                    raise ValueError(
                        'Cannot use ssl_verify=True with http host')
            values['ssl_verify'] = ssl_verify
            return server
        elif isinstance(url, NetboxServerModel):
            return url
        else:
            raise ValueError('Unknown input type')

    @validator('token')
    def validate_token(cls, token):
        """checks if the token can be load as sensible data
        """
        try:
            if token == 'ask':
                return token
            return get_sensitive_data(token)
        except SensitiveLoadError as e:
            raise ValueError(e)


class Netbox(Source, InventoryAsyncPlugin):
    """This class is used to dinamically retrieve the inventory from Netbox
       and also retrieve for each device in the inventory its credentials

    """

    def __init__(self, config_data: dict, validate: bool = True) -> None:
        self._status = 'init'
        self._session: aiohttp.ClientSession = None
        self._server: NetboxServerModel = None

        super().__init__(config_data, validate)

    @classmethod
    def get_data_model(cls):
        return NetboxSourceModel

    def _load(self, input_data):
        # load the server class from the dictionary
        if not self._validate:
            input_data['server'] = NetboxServerModel.construct(
                **input_data.pop('url', {}))
            input_data['ssl_verify'] = input_data.pop('ssl-verify', False)
        super()._load(input_data)
        if self._data.token == 'ask':
            self._data.token = get_sensitive_data(
                'ask', f'{self.name} Insert netbox API token: '
            )
        self._server = self._data.server
        if not self._auth:
            raise InventorySourceError(f"{self.name} Netbox must have an "
                                       "'auth' set in the 'namespaces' section"
                                       )

    def _init_session(self, headers: dict):
        """Initialize the session property

        Args:
            headers ([dict]): headers to initialize the session
        """
        if not self._session:
            ssl_option = None if self._data.ssl_verify else False
            self._session = aiohttp.ClientSession(
                headers=headers,
                connector=aiohttp.TCPConnector(ssl=ssl_option)
            )

    def _token_auth_header(self) -> Dict:
        """Generate the token authorization header

        Returns:
            Dict: token authorization header
        """
        return {'Authorization': f'Token {self._data.token}'}

    async def get_inventory_list(self) -> List:
        """Contact netbox to retrieve the inventory.

        Raises:
            RuntimeError: Unable to connect to the REST server

        Returns:
            List: inventory list
        """
        url = f'{self._server.protocol}://{self._server.host}:'\
            f'{self._server.port}/api/dcim/devices/?tag={self._data.tag}'
        if not self._session:
            headers = self._token_auth_header()
            self._init_session(headers)
        try:
            devices, next_url = await self._get_devices(url)
            while next_url:
                cur_devices, next_url = await self._get_devices(next_url)
                devices.extend(cur_devices)
        except Exception as e:
            raise InventorySourceError(f'{self.name}: error while '
                                       f'getting devices: {e}')

        logger.info(
            f'Netbox: Retrieved inventory list of {len(devices)} devices')
        return devices

    async def _get_devices(self, url: str) -> Tuple[List, str]:
        """Retrieve devices from netbox using an HTTP GET over <url>

        Args:
            url (str): devices url

        Raises:
            RuntimeError: Response error

        Returns:
            Tuple[List, str]: returns the list of devices and the url to get
            the remaining devices
        """
        async with self._session.get(url) as response:
            if int(response.status) == 200:
                res = await response.json()
                return res.get('results', []), res.get('next')
            else:
                raise InventorySourceError(
                    f'{self.name}: error in inventory get '
                    f'{await response.text()}')

    def parse_inventory(self, inventory_list: list) -> Dict:
        """parse the raw inventory collected from the server and generates
           a new inventory with only the required informations

        Args:
            raw_inventory (list): raw inventory received from the server

        Returns:
            List[Dict]: a list containing the inventory
        """
        def get_field_value(entry: dict, fields_str: str):
            # Fields_str is a string containing fields separated by dots
            # Every dot means that the value is deeper in the dictionary
            # Example: fields_str = 'site.name'
            #          return entry[site][name]
            fields = fields_str.split('.')
            cur_field = None
            for i, field in enumerate(fields):
                if i == 0:
                    cur_field = entry.get(field, None)
                else:
                    cur_field = cur_field.get(field, None)
                if cur_field is None:
                    return None
            return cur_field

        inventory = {}
        ignored_device_count = 0

        for device in inventory_list:
            ipv4 = get_field_value(device, 'primary_ip4.address')
            ipv6 = get_field_value(device, 'primary_ip6.address')
            hostname = get_field_value(device, 'name')
            site_name = get_field_value(device, 'site.name')
            if self._namespace == 'netbox-sitename':
                namespace = site_name
            else:
                namespace = self._namespace
            address = None

            if ipv4:
                address = ipv4
            elif ipv6:
                address = ipv6

            if not address:
                logger.warning(
                    f"Skipping {namespace}.{hostname}: doesn't have a "
                    "management IP")
                ignored_device_count += 1
                continue

            address = address.split("/")[0]

            inventory[f'{namespace}.{address}'] = {
                'address': address,
                'namespace': namespace,
                'hostname': hostname,
            }

        logger.info(
            f'Netbox: Acting on inventory of {len(inventory)} devices, '
            f'ignoring {ignored_device_count} devices')

        return inventory

    async def _execute(self):
        while True:
            inventory_list = await self.get_inventory_list()
            logger.debug(
                f"Received netbox inventory from {self._server.protocol}"
                f"://{self._server.host}:{self._server.port}")
            tmp_inventory = self.parse_inventory(inventory_list)
            # Write the inventory and remove the tmp one
            self.set_inventory(tmp_inventory)

            if self._run_once:
                break

            await asyncio.sleep(self._data.period)

    async def _stop(self):
        if self._session:
            await self._session.close()



from suzieq.sqobjects.basicobj import SqObject


class IfCountersObj(SqObject):
    '''The object providing access to the interface counters table'''

    def __init__(self, **kwargs):
        super().__init__(table='ifCounters', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'columns', 'ifname',
                                'query_str']



import time
import ipaddress
import pandas as pd

from nubia import command
from suzieq.cli.nubia_patch import argument

from suzieq.cli.sqcmds.command import SqTableCommand
from suzieq.sqobjects.routes import RoutesObj


@command("route", help="Act on Routes")
@argument("vrf", description="VRF(s), space separated")
@argument("protocol", description="Routing protocol(s), space separated")
@argument("prefix", description="Prefix(es), in quotes, space separated")
@argument("prefixlen", description="must be of the form "
          "[<|<=|>=|>|!] length")
class RouteCmd(SqTableCommand):
    """Routing table information"""

    def __init__(
            self,
            engine: str = "",
            hostname: str = "",
            start_time: str = "",
            end_time: str = "",
            view: str = "",
            namespace: str = "",
            format: str = "",  # pylint: disable=redefined-builtin
            query_str: str = ' ',
            columns: str = "default",
            vrf: str = "",
            protocol: str = "",
            prefix: str = '',
            prefixlen: str = '',
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=RoutesObj,
        )
        self.lvars = {
            'protocol': protocol.split(),
            'vrf': vrf.split(),
            'prefix': prefix.split(),
            'prefixlen': prefixlen
        }

    def _json_print_handler(self, in_data):  # pylint: disable=method-hidden
        """This handler calls the code to print the IPNetwork as a string"""
        if isinstance(in_data, ipaddress.IPv4Network):
            return ipaddress.IPv4Network.__str__(in_data)
        elif isinstance(in_data, ipaddress.IPv6Network):
            return ipaddress.IPv6Network.__str__(in_data)
        return in_data

    def _get_ipvers(self, value: str) -> int:
        """Return the IP version in use"""

        if ':' in value:
            ipvers = 6
        elif '.' in value:
            ipvers = 4
        else:
            ipvers = ''

        return ipvers

    @command('lpm')
    @argument("address", description="IP Address, in quotes, for lpm query")
    def lpm(self, address: str = ''):
        """Show the Longest Prefix Match(LPM) on a given prefix, vrf
        """
        now = time.time()
        if self.columns != ["default"]:
            self.ctxt.sort_fields = None
        else:
            self.ctxt.sort_fields = []

        if not address:
            df = pd.DataFrame({'error': ['Must specify address for lpm']})
            return self._gen_output(df)

        df = self._invoke_sqobj(self.sqobj.lpm,
                                hostname=self.hostname,
                                address=address,
                                ipvers=self._get_ipvers(address),
                                columns=self.columns,
                                namespace=self.namespace,
                                query_str=self.query_str,
                                **self.lvars
                                )

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        return self._gen_output(df)



from typing import List
from ipaddress import IPv4Network

import numpy as np
import pandas as pd

from suzieq.engines.pandas.engineobj import SqPandasEngine
from suzieq.shared.utils import build_query_str, humanize_timestamp
from suzieq.shared.schema import SchemaForTable


class OspfObj(SqPandasEngine):
    '''Backend class to handle manipulating OSPF table with pandas'''

    @staticmethod
    def table_name():
        '''Table name'''
        return 'ospf'

    # pylint: disable=too-many-statements
    def _get_combined_df(self, **kwargs):
        """OSPF has info divided across multiple tables. Get a single one"""

        columns = kwargs.pop('columns', ['default'])
        state = kwargs.pop('state', '')
        user_query = kwargs.pop('query_str', '')
        hostname = kwargs.pop('hostname', [])

        cols = self.schema.get_display_fields(columns)
        self._add_active_to_fields(kwargs.get('view', self.iobj.view), cols,
                                   None)

        user_query_cols = self._get_user_query_cols(user_query)

        ifschema = SchemaForTable('ospfIf', schema=self.all_schemas)
        nbrschema = SchemaForTable('ospfNbr', schema=self.all_schemas)

        if columns not in [['default'], ['*']]:
            ifkeys = ifschema.key_fields()
            nbrkeys = nbrschema.key_fields()
            if_flds = ifschema.fields
            nbr_flds = nbrschema.fields

            ifcols = ifkeys
            nbrcols = nbrkeys
            for fld in columns:
                if fld in if_flds and fld not in ifcols:
                    ifcols.append(fld)
                elif fld in nbr_flds and fld not in nbrcols:
                    nbrcols.append(fld)
                if 'state' not in nbrcols:
                    nbrcols.append('state')
                ifcols += [x for x in ['area', 'state', 'passive']
                           if x not in ifcols]
        else:
            ifcols = ifschema.get_display_fields(columns)
            nbrcols = nbrschema.get_display_fields(columns)

        ifcols += [x for x in user_query_cols if (x in ifschema.fields
                                                  and x not in ifcols)]
        self._add_active_to_fields(kwargs.get('view', 'latest'), ifcols,
                                   None)
        nbrcols += [x for x in user_query_cols if (x in nbrschema.fields
                                                   and x not in nbrcols)]
        self._add_active_to_fields(kwargs.get('view', 'latest'), nbrcols,
                                   None)
        state_query_dict = {
            'full': '(adjState == "full" or adjState == "passive")',
            'passive': '(adjState == "passive")',
            'other': '(adjState != "full" and adjState != "passive")',
            '!full': '(adjState != "full")',
            '!passive': '(adjState != "passive")',
            '!other': '(adjState == "full" or adjState == "passive")',
        }

        if state:
            query_str = state_query_dict.get(state, '')
            cond_prefix = ' and '
        else:
            query_str = ''
            cond_prefix = ''

        host_query_str = build_query_str([], ifschema, ignore_regex=False,
                                         hostname=hostname)
        if host_query_str:
            query_str += f'{cond_prefix}{host_query_str}'

        df = self._get_table_sqobj('ospfIf') \
                 .get(columns=ifcols, **kwargs)
        nbr_df = self._get_table_sqobj('ospfNbr') \
                     .get(columns=nbrcols, **kwargs)
        if nbr_df.empty:
            return df

        merge_cols = [x for x in ['namespace', 'hostname', 'ifname']
                      if x in nbr_df.columns]
        # Merge the two tables
        df = df.merge(nbr_df, on=merge_cols, how='left') \
               .fillna({'peerIP': '-', 'numChanges': 0,
                        'lastChangeTime': 0}) \
               .fillna('')

        # This is because some NOS have the ipAddress in nbr table and some in
        # interface table. Nbr table wins over interface table if present
        if 'ipAddress_y' in df:
            df['ipAddress'] = np.where(
                df['ipAddress_x'] == "",
                df['ipAddress_y'], df['ipAddress_x'])

        if columns == ['*']:
            df = df.drop(columns=['area_y', 'instance_y', 'vrf_y',
                                  'ipAddress_x', 'ipAddress_y', 'areaStub_y',
                                  'sqvers_x', 'timestamp_y'],
                         errors='ignore') \
                .rename(columns={
                    'instance_x': 'instance', 'areaStub_x': 'areaStub',
                    'area_x': 'area', 'vrf_x': 'vrf',
                    'state_x': 'ifState', 'state_y': 'adjState',
                    'active_x': 'active', 'timestamp_x': 'timestamp'}) \
                .fillna({'peerIP': '-', 'numChanges': 0,
                         'lastChangeTime': 0})
        else:
            df = df.rename(columns={'vrf_x': 'vrf', 'area_x': 'area',
                                    'state_x': 'ifState',
                                    'state_y': 'adjState',
                                    'timestamp_x': 'timestamp'})
            df = df.drop(list(df.filter(regex='_y$')), axis=1) \
                   .drop(columns=['ipAddress_x'], errors='ignore') \
                   .fillna({'peerIP': '-', 'numChanges': 0,
                            'lastChangeTime': 0})
        if df.empty:
            return df

        if 'state' in df.columns:
            # Need this logic if the user specfies only one of adjState
            # or ifState in the columns field. The above renaming of state_x
            # and state_y will not work in that case, and this is what we
            # have to do.
            if 'ifState' in cols:
                df = df.rename(columns={'state': 'ifState'})
            else:
                df = df.rename(columns={'state': 'adjState'})
        if 'lastChangeTime' in df.columns:
            df['lastChangeTime'] = np.where(df.lastChangeTime == '-',
                                            0, df.lastChangeTime)
        # Fill the adjState column with passive if passive
        if 'passive' in df.columns and 'adjState' in df.columns:
            df.loc[df['adjState'] == '', 'adjState'] = df['passive']
            df.loc[df['adjState'].eq(True), 'adjState'] = 'passive'
            df.loc[df['adjState'].eq(False), 'adjState'] = 'fail'
            df.loc[df['adjState'] == 'passive', 'peerIP'] = ''
            df.loc[df['adjState'] == 'passive', 'peerRouterId'] = ''

            df.drop(columns=['passive'], inplace=True)

        final_df = df
        if 'active_x' in final_df.columns:
            final_df = final_df.rename(columns={'active_x': 'active'})
        if 'peerHostname' in cols or 'peerIfname' in cols:
            final_df = self._get_peernames(final_df, cols, hostname=hostname,
                                           **kwargs)
        if query_str:
            final_df = final_df.query(query_str).reset_index(drop=True)

        if user_query and not final_df.empty:
            final_df = self._handle_user_query_str(final_df, user_query)
        # Move the timestamp column to the end

        return final_df.reset_index(drop=True)[cols]

    def get(self, **kwargs):
        return self._get_combined_df(**kwargs)

    def summarize(self, **kwargs):
        """Describe the data"""

        # Discard these
        kwargs.pop('columns', None)

        # 'ospfIf' is ignored
        self._init_summarize(**kwargs)
        if self.summary_df.empty:
            return self.summary_df

        self._summarize_on_add_field = [
            ('deviceCnt', 'hostname', 'nunique'),
            ('peerCnt', 'hostname', 'count'),
        ]

        self._summarize_on_add_with_query = [
            ('stubbyPeerCnt', 'areaStub', 'areaStub'),
            ('passivePeerCnt', 'adjState == "passive"', 'ifname'),
            ('unnumberedPeerCnt', 'isUnnumbered', 'isUnnumbered'),
            ('failedPeerCnt', '~adjState.isin(["passive", "full"])',
             'ifname'),
        ]

        self._summarize_on_add_list_or_count = [
            ('area', 'area'),
            ('vrf', 'vrf'),
            ('helloTime', 'helloTime'),
            ('deadTime', 'deadTime'),
            ('retxTime', 'retxTime'),
            ('networkType', 'networkType'),
        ]

        self.summary_df['lastChangeTime'] = np.where(
            self.summary_df.lastChangeTime.isnull(), 0,
            self.summary_df.lastChangeTime)

        self.summary_df['lastChangeTime'] = humanize_timestamp(
            self.summary_df.lastChangeTime, self.cfg.get('analyzer', {})
            .get('timezone', None))

        self.summary_df['lastChangeTime'] = (
            self.summary_df['timestamp'] - self.summary_df['lastChangeTime'])
        self.summary_df['lastChangeTime'] = self.summary_df['lastChangeTime'] \
            .apply(lambda x: x.round('s'))

        self._summarize_on_add_stat = [
            ('adjChangesStat', '', 'numChanges'),
            ('upTimeStat', 'adjState == "full"', 'lastChangeTime'),
        ]

        self._gen_summarize_data()
        self._post_summarize()
        return self.ns_df.convert_dtypes()

    def aver(self, **kwargs):
        """Assert that the OSPF state is OK"""

        kwargs.pop('columns', [])
        columns = ['*']

        result = kwargs.pop('result', 'all')
        kwargs.pop('state', '')
        # we have to not filter hostname at this point because we need to
        #   understand neighbor relationships

        ospf_df = self.get(columns=columns, **kwargs)

        if not ospf_df.empty:
            ospf_df = ospf_df.query('ifState != "adminDown"') \
                             .reset_index(drop=True)

        if ospf_df.empty:
            return pd.DataFrame(columns=columns)

        columns.extend(['peerHostname', 'peerIfname'])
        ospf_df = self._get_peernames(ospf_df, columns)

        ospf_df["assertReason"] = [[]]*ospf_df.shape[0]
        df = (
            ospf_df[ospf_df["routerId"] != ""]
            .groupby(["routerId", "namespace"], as_index=False)[["hostname",
                                                                 "namespace"]]
            .agg(lambda x: x.unique().tolist())
        ).dropna(how='any')

        # df is a dataframe with each row containing the routerId and the
        # corresponding list of hostnames with that routerId. In a good
        # configuration, the list must have exactly one entry.
        if not df.empty:
            # This check is because we don't get routerID with Arista boxes
            ospf_df['assertReason'] = (
                ospf_df.merge(df, on=["routerId"], how="outer")
                .apply(lambda x: ["duplicate routerId {}".format(
                    x["hostname_y"])]
                    if (x["hostname_y"] is not np.nan and
                        len(x['hostname_y']) != 1) else [], axis=1))

        ok_df = ospf_df \
            .query('adjState.isin(["passive", "full"])') \
            .reset_index(drop=True)
        failed_df = ospf_df \
            .query('~adjState.isin(["passive", "full"]) and '
                   'ifState != "down"') \
            .reset_index(drop=True)
        ifdown_df = ospf_df \
            .query('~adjState.isin(["full", "passive"]) and '
                   'ifState == "down"') \
            .reset_index(drop=True)
        ifdown_df['assertReason'] += \
            ifdown_df.ifState.apply(lambda x: ['ifdown'])

        # Validate loopback interfaces are of the right network type
        if not ok_df.empty:
            ok_df['assertReason'] += ok_df.apply(
                lambda x: ['network type !loopback']
                if (any(ele in x.ifname for ele in ['lo', 'oopbac']) and
                    x.networkType != "loopback") else [], axis=1)

        if failed_df.empty:
            return self._create_aver_result([ok_df, ifdown_df], result)

        # We have the MTU mismatch which requires us to pull the MTU from the
        # interfaces table
        ifdf = self._get_table_sqobj('interfaces').get(
            namespace=kwargs.get('namespace', []),
            hostname=kwargs.get('hostname', []),
            columns=['namespace', 'hostname', 'ifname', 'mtu',
                     'ipAddressList'],
            query_str='ipAddressList.str.len() != 0')

        if not ifdf.empty:
            ifdf = ifdf.explode('ipAddressList')
            ifdf['ipAddressList'] = ifdf.ipAddressList.str.split('/').str[0]
            failed_df = failed_df.merge(
                ifdf, how='left',
                on=['namespace', 'hostname', 'ifname'],
                suffixes=('', '_y'))
            check_mtu = True
        else:
            check_mtu = False

        peer_df = failed_df.merge(
            failed_df, how='left',
            left_on=["namespace", "hostname", "ifname"],
            right_on=["namespace", "peerHostname", "peerIfname"]) \
            .rename(columns={'assertReason_x': 'assertReason'}) \
            .dropna(subset=['ipAddress_x', 'index_x', 'ipAddress_y']) \
            .fillna({'isUnnumbered_y': False})

        if peer_df.empty:
            # We're unable to find a peer but the sessions have failed
            failed_df['assertReason'] = failed_df.apply(
                lambda x: 'Loopback not configured passive'
                if x['networkType'] == 'loopback' else 'No Peer Found',
                axis=1)

            return self._create_aver_result([ok_df, ifdown_df, failed_df],
                                            result)

        peer_df = peer_df.drop(columns=['assertReason_y'])

        # Now start comparing the various parameters
        # pylint: disable=condition-evals-to-constant
        peer_df["assertReason"] += peer_df.apply(
            lambda x: ["subnet mismatch"]
            if (
                (x["isUnnumbered_x"] != x["isUnnumbered_y"])
                or (
                    (('ipAddress_y == ""') or
                     ((IPv4Network(x["ipAddress_x"], strict=False)
                       != IPv4Network(x["ipAddress_y"], strict=False)))))
            )
            else [],
            axis=1,
        )

        peer_df["assertReason"] += peer_df.apply(
            lambda x: ["area mismatch"]
            if (x["area_x"] != x["area_y"]) else [], axis=1)

        peer_df["assertReason"] += peer_df.apply(
            lambda x: ["area stub mismatch"]
            if (x["areaStub_x"] != x["areaStub_y"]) else [], axis=1)

        peer_df["assertReason"] += peer_df.apply(
            lambda x: ["Hello timers mismatch"]
            if x["helloTime_x"] != x["helloTime_y"]
            else [],
            axis=1,
        )
        peer_df["assertReason"] += peer_df.apply(
            lambda x: ["Dead timer mismatch"]
            if x["deadTime_x"] != x["deadTime_y"]
            else [],
            axis=1,
        )
        peer_df["assertReason"] += peer_df.apply(
            lambda x: ["network type mismatch"]
            if x["networkType_x"] != x["networkType_y"]
            else [],
            axis=1,
        )
        peer_df["assertReason"] += peer_df.apply(
            lambda x: ["vrf mismatch"] if x["vrf_x"] != x["vrf_y"] else [],
            axis=1,
        )

        if check_mtu:
            peer_df["assertReason"] += peer_df.apply(
                lambda x: ["MTU mismatch"]
                if x["mtu_x"] != x["mtu_y"] else [],
                axis=1,
            )

        # Now walk through and label all the failed sessions for
        # which we have no reason
        peer_df['assertReason'] += peer_df.apply(
            lambda x: ["No Peer Found"]
            if (x['adjState_x'] == "fail" and
                x['assertReason'].str.len() == 0)
            else [],
            axis=1)

        result_df = (
            peer_df.rename(
                index=str,
                columns={
                    "hostname_x": "hostname",
                    "ifname_x": "ifname",
                    "vrf_x": "vrf",
                    "timestamp_x": "timestamp",
                    "adjState_x": "adjState",
                },
            )[["namespace", "hostname", "ifname", "vrf", "adjState",
               "assertReason", "timestamp"]]
        )

        return self._create_aver_result([ok_df, ifdown_df, result_df],
                                        result)

    def _get_peernames(self, df: pd.DataFrame, columns: List[str],
                       **kwargs) -> pd.DataFrame:
        """Fill in columns of peerhostname and peerifname if requested

        Using a combination of LLDP info, if available and without it,
        this routine tries to fill in the columns of peerhostname and
        peerifname for the relevant rows. In the absence of LLDP info,
        filling in peerifname is very error-prone especially in the
        presence of unnumbered interfaces.

        Args:
            df: The OSPF dataframe with the relevant dependent cols
            columns: List of column names provided by user
            kwargs: User specified list for hostname/namespace etc.

        Returns:
            The OSPF dataframe with the peerHostname and peerifname
            filled in
        """

        if df.empty or not any(x in columns
                               for x in ['peerHostname', 'peerIfname', '*',
                                         'default']):
            return df

        df['peerHostname'] = ''
        df['peerIfname'] = ''

        nfdf = df.query('adjState == "passive"').reset_index()
        newdf = df.query('adjState != "passive"').reset_index() \
            .drop('peerHostname', axis=1, errors='ignore')
        if newdf.empty:
            return df

        lldp_df = self._get_table_sqobj('lldp').get(
            namespace=kwargs.get('namespace', []),
            hostname=kwargs.get('hostname', []),
            use_bond='True')

        if lldp_df.empty:
            # Try a match entirely within ospf, but we can't
            # get ifname effectively in cases of unnumbered
            newdf['matchIP'] = newdf.ipAddress.str.split('/').str[0]
            newdf = newdf \
                .merge(newdf[['namespace', 'hostname', 'vrf',
                              'matchIP']],
                       left_on=['namespace', 'vrf', 'peerIP'],
                       right_on=['namespace', 'vrf', 'matchIP'],
                       suffixes=["", "_y"]) \
                .rename(columns={'hostname_y': 'peerHostname'}) \
                .drop_duplicates(subset=['namespace', 'hostname',
                                         'vrf', 'ifname']) \
                .drop(columns=['matchIP', 'matchIP_y', 'timestamp_y'],
                      errors='ignore')
            if 'peerIfname' in columns:
                newdf['peerIfname'] = ''
        else:
            # In case of Junos, the OSPF interface name is a subif of
            # the interface name. So, create a new column with the
            # orignal interface name.
            newdf['lldpIfname'] = newdf['ifname'].str.split('.').str[0]

            newdf = newdf.merge(
                lldp_df[['namespace', 'hostname', 'ifname',
                         'peerHostname', 'peerIfname']],
                left_on=['namespace', 'hostname', 'lldpIfname'],
                right_on=['namespace', 'hostname', 'ifname'],
                how='outer', suffixes=('', '_y')) \
                .drop(columns=['ifname_y', 'lldpIfname', 'peerIfname']) \
                .rename(columns={'peerIfname_y': 'peerIfname'}) \
                .dropna(subset=['hostname', 'vrf', 'ifname'], how='any')\
                .fillna('')

        if newdf.empty:
            newdf = df.query('adjState != "passive"').reset_index(drop=True)
            newdf['peerHostname'] = ''

        # Lets find the rows without peerHostnames
        nopeer_df = newdf.query('peerHostname == ""')
        if not nopeer_df.empty:
            peerdf = newdf.query('peerHostname != ""')
            nopeer_df = nopeer_df.drop(columns=['peerHostname', 'peerIfname'],
                                       errors='ignore')
            # We need to look deeper to see if we can figure out the peers
            addr_df = self._get_table_sqobj('address').get(
                namespace=kwargs.get('namespace', []),
                hostname=kwargs.get('hostname', []),
                type='!loopback',
                columns=['namespace', 'hostname', 'vrf', 'ifname', 'state',
                         'ipAddressList'],
                query_str='ipAddressList.str.len() != 0')
            if not addr_df.empty:
                addr_df = addr_df.explode('ipAddressList')
                addr_df['ipAddressList'] = addr_df.ipAddressList \
                                                  .str.split('/').str[0]
                nopeer_df = nopeer_df.merge(
                    addr_df,
                    left_on=['namespace', 'peerIP'],
                    right_on=['namespace', 'ipAddressList'],
                    suffixes=['', '_y'], how='left') \
                    .rename(columns={'hostname_y': 'peerHostname',
                                     'ifname_y': 'peerIfname'}) \
                    .drop(columns=['state', 'ipAddressList'],
                          errors='ignore') \
                    .fillna('')

        if not nopeer_df.empty:
            final_df = pd.concat([nfdf, peerdf, nopeer_df])
        else:
            final_df = pd.concat([nfdf, newdf])

        return final_df

    def _create_aver_result(self, dflist: List[pd.DataFrame],
                            result: str) -> pd.DataFrame:
        """Create a result dataframe from list and return a single df

        A common function to concat all the different dataframes
        created during assert such as passed_df, failed_df etc.,
        create a single dataframe and return the list given user input
        to return all rows or just some

        Args:
            dflist: List[pd.DataFrame]
            result: str: user specification of what result to return

        Returns:
            The assert result dataframe
        """
        ospf_df = pd.concat(dflist)
        ospf_df['result'] = ospf_df.assertReason.apply(
            lambda x: 'fail' if x else 'pass')
        # Weed out the loopback, SVI interfaces as they have no LLDP peers
        if result and result != "all":
            ospf_df = ospf_df.query(f'result == "{result}"')

        return ospf_df[['namespace', 'hostname', 'vrf', 'ifname', 'adjState',
                        'assertReason', 'result']] \
            .explode('assertReason') \
            .fillna('') \
            .reset_index(drop=True)



'''
Device simulator that serves up fake SSH/REST server

The command outputs are reflected from files with the appropriate
command name and format extension. show_lldp_neighbors_detail.txt,
for example
'''
import sys
import subprocess
import argparse
from pathlib import Path

import yaml
from jsonschema import validate
from jsonschema.exceptions import ValidationError

from suzieq.shared.utils import get_sq_install_dir

schema = {
    "type": "object",
    "properties": {
        "devices": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "input-dir": {
                        "type": "string"
                    },
                    "hostname": {
                        "type": "string"
                    },
                    "ssh-port": {
                        "type": "integer"
                    },
                    "rest-port": {
                        "type": "integer"
                    },
                },
                "required": [
                    "input-dir", "hostname"
                ]
            }
        }
    }
}


# pylint: disable=redefined-outer-name
def run_simnodes(devices: list = None, input_dir: str = None,
                 port: int = None, transport: str = None):
    """Start ssh and rest simnode subprocesses with the options
    (nos, version, hostname) given in the configuration file
    """
    processes = []

    for dev in devices:
        topdir = Path(dev['input-dir'])
        if not topdir.exists():
            print(f'ERROR: Unable to find datadir {input_dir}')
            sys.exit(1)

        if dev.get('hostname', '') == "all":
            hostdirs = [x for x in topdir.iterdir() if x.is_dir()]
        else:
            hostdirs = [Path(topdir) / dev['hostname']]

        ssh_port = dev.get("ssh-port", 0)
        rest_port = dev.get("rest-port", 0)

        processes += run_simnodes_dir(hostdirs, ssh_port, rest_port)

    if input_dir:
        hostdirs = [x for x in input_dir.iterdir() if x.is_dir()]
        if transport == "ssh":
            ssh_port = port
            rest_port = 0
        else:
            rest_port = port
            ssh_port = 0

        processes += run_simnodes_dir(hostdirs, ssh_port, rest_port)

    for proc in processes:
        proc.wait()


def run_simnodes_dir(hostdirs: str, base_ssh_port: int,
                     base_rest_port: int) -> list:
    '''Run one or more sessions given the dir and port'''
    processes = []

    bindir = f'{get_sq_install_dir()}/utilities'

    for i, hostd in enumerate(sorted(hostdirs)):
        if base_ssh_port:
            ssh_port = base_ssh_port + i
            cmd = [
                "python", f"{bindir}/sq_simnode_ssh.py",
                "-p", str(ssh_port),
                "-d", hostd,
            ]
            # pylint: disable=consider-using-with
            proc = subprocess.Popen(cmd)
            print(
                f'Running ssh node with input dir {hostd} '
                f'for {hostd.parts[-1]} on port {ssh_port}')
            processes.append(proc)

        if base_rest_port:
            rest_port = base_rest_port + i
            cmd = [
                "python", f"{bindir}/sq_simnode_rest.py",
                "-p", str(rest_port),
                "-d", hostd,
            ]
            # pylint: disable=consider-using-with
            proc = subprocess.Popen(cmd)
            print(
                f'Running REST node with input dir {hostd} '
                f'for {hostd.parts[-1]} on port {rest_port}')

            processes.append(proc)

    return processes


def simnodes_main():
    '''The main routine'''
    parser = argparse.ArgumentParser()
    inputs = parser.add_mutually_exclusive_group(required=True)
    inputs.add_argument(
        "-D", "--devices-file", type=str, help="Device list file",
    )
    inputs.add_argument(
        "-i", "--input-dir", type=str, help="Directory containing host files"
    )
    parser.add_argument(
        "-p", "--port", type=int, help="Starting port to listen on",
        default=10000
    )
    parser.add_argument(
        "-t", "--transport", type=str, help="Starting port to listen on",
        default="ssh", choices=["ssh", "rest"]
    )

    args = parser.parse_args()

    devs = {}

    if args.devices_file:
        with open(args.devices_file, encoding='utf8') as fh:
            devs = yaml.load(fh, Loader=yaml.FullLoader)

        try:
            validate(devs, schema)
        except ValidationError as e:
            print(e.message)
            sys.exit(1)
    else:
        input_dir = Path(args.input_dir)
        if not input_dir.exists():
            print(f'ERROR: Input dir {args.input_dir} does not exist')
            sys.exit(1)
        if not input_dir.is_dir():
            print(f'ERROR: Input dir {args.input_dir} is not a dir')
            sys.exit(1)

    run_simnodes(devs, input_dir=input_dir, port=args.port,
                 transport=args.transport)


if __name__ == '__main__':
    simnodes_main()



"""
This module contains the logic of the writer for the 'gather' mode
"""
import os

from suzieq.poller.worker.writers.output_worker import OutputWorker


class GatherOutputWorker(OutputWorker):
    """GatherOutputWorker is used to write poller output
    in the case of the run-once data gather mode
    """

    def write_data(self, data):
        """Write the output of the commands into a plain file

        Args:
            data (Dict): dictionary containing the data to store.
        """
        file = os.path.join(self.root_output_dir, f"{data['topic']}.output")
        with open(file, 'a') as f:
            # Even though we use JSON dump, the output is not valid JSON
            f.write(data['records'])



# pylint: disable=no-name-in-module

from typing import Dict, List, Optional
from urllib3.util import parse_url
from pydantic import BaseModel, Field, validator

from suzieq.shared.utils import PollerTransport


class InventoryModel(BaseModel):
    """Model for the inventory validation
    """
    sources: List[Dict] = Field(min_items=1)
    namespaces: List[Dict] = Field(min_items=1)
    auths: Optional[List[Dict]]
    devices: Optional[List[Dict]]

    class Config:
        """pydantic configuration
        """
        extra = 'forbid'


class DeviceModel(BaseModel):
    """Device model validation
    """
    name: str
    jump_host: Optional[str] = Field(alias='jump-host')
    jump_host_key_file: Optional[str] = Field(alias='jump-host-key-file')
    ignore_known_hosts: Optional[bool] = Field(
        alias='ignore-known-hosts', default=False)
    slow_host: Optional[bool] = Field(alias='slow-host', default=False)
    per_cmd_auth: Optional[bool] = Field(alias='per-cmd-auth', default=True)
    retries_on_auth_fail: Optional[int] = Field(alias='retries-on-auth-fail',
                                                default=1)
    transport: Optional[PollerTransport]
    port: Optional[str]
    devtype: Optional[str]

    # pylint: disable=no-self-argument
    @validator('jump_host')
    def jump_host_validator(cls, value):
        '''Validate jump host format'''
        try:
            uri = parse_url(value)
            assert uri.scheme is None, \
                'format username@jumphost[:port] required'
        except ValueError:
            assert False, f'Invalid jumphost format: {value}'

    class Config:
        """pydantic configuration
        """
        extra = 'forbid'


class NamespaceModel(BaseModel):
    """Namespace model validation
    """
    name: str
    source: str
    device: Optional[str]
    auth: Optional[str]

    class Config:
        """pydantic configuration
        """
        extra = 'forbid'



from nubia import command

from suzieq.cli.sqcmds.command import SqTableCommand
from suzieq.sqobjects.topcpu import TopcpuObj


@command("topcpu", help="Act on topcpu data")
class TopcpuCmd(SqTableCommand):
    """Information about Top CPU users"""

    def __init__(
        self,
        engine: str = "",
        hostname: str = "",
        start_time: str = "",
        end_time: str = "",
        view: str = "",
        namespace: str = "",
        format: str = "",  # pylint: disable=redefined-builtin
        columns: str = "default",
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            sqobj=TopcpuObj
        )



import time

from nubia import command
import pandas as pd

from suzieq.cli.nubia_patch import argument
from suzieq.sqobjects import get_sqobject
from suzieq.cli.sqcmds.command import SqTableCommand


@command("interface", help="Act on Interface data")
@argument("ifname", description="interface name(s), space separated")
@argument("type", description="interface type(s), space separated")
@argument("vrf", description="VRF(s), space separated")
@argument("portmode", description="Portmode(s), space separated")
@argument("vlan", description="Vlan(s), space separated")
@argument("vrf", description="VRF(s), space separated")
@argument("state", description="interface state to qualify show",
          choices=["up", "down", "notConnected", "!up", "!down",
                   "!notConnected"])
@argument("mtu",
          description="MTU(s), space separated, can use <, >, <=, >=, !")
class InterfaceCmd(SqTableCommand):
    """Device interface information including MTU, Speed, IP address etc"""

    # pylint: disable=redefined-builtin
    def __init__(
            self,
            engine: str = "",
            hostname: str = "",
            start_time: str = "",
            end_time: str = "",
            view: str = "",
            namespace: str = "",
            format: str = "",  # pylint: disable=redefined-builtin
            columns: str = "default",
            query_str: str = ' ',
            ifname: str = '',
            state: str = '',
            type: str = '',
            vrf: str = '',
            mtu: str = '',
            portmode: str = '',
            vlan: str = '',
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=get_sqobject('interfaces')
        )
        self.lvars = {
            'ifname': ifname.split(),
            'state': state,
            'type': type.split(),
            'vrf': vrf.split(),
            'mtu': mtu.split(),
            'vlan': vlan.split(),
            'portmode': portmode.split(),
        }

    @command("assert")
    @argument(
        "what",
        description="What do you want to assert",
        choices=["mtu-value"],
    )
    @argument("value", description="Value to match against")
    @argument("result", description="Show only assert that matches this value",
              choices=["all", "fail", "pass"])
    @argument("ignore_missing_peer",
              description="Treat missing peer as passing assert check",
              choices=["True", "False"])
    def aver(self, what: str = "",
             value: str = '', result: str = 'all',
             ignore_missing_peer: str = "False"):
        """Assert aspects about the interface
        """

        now = time.time()

        if what == "mtu-value" and value == '':
            df = pd.DataFrame(
                {'error': ["Provide value to match MTU against"]}
            )
            return self._gen_output(df)

        elif not what:
            value = ''

        df = self._invoke_sqobj(self.sqobj.aver,
                                hostname=self.hostname,
                                namespace=self.namespace,
                                what=what,
                                matchval=value.split(),
                                result=result,
                                ignore_missing_peer=(
                                    ignore_missing_peer == "True"),
                                **self.lvars,
                                )

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)

        return self._assert_gen_output(df)



'''
Serve up a fake REST server acting as device REST API
'''
import sys
import argparse
from pathlib import Path
import ssl
import logging

from aiohttp import web
import xmltodict


def get_filename_from_cmd(cmd_dict: dict) -> str:
    """Build filename from an xml command"""

    keys = []

    def recursive_items(cmd_dict: dict):
        for key, value in cmd_dict.items():
            if isinstance(value, dict):
                # do not use list entry for filename
                if key != "entry":
                    keys.append(key)
                recursive_items(value)
            elif value:
                # do not use property names for filename
                if not key.startswith("@"):
                    keys.append(key)
                keys.append(value)
            else:
                keys.append(key)

    recursive_items(cmd_dict)

    return "_".join(keys).replace("-", "_")


def run_server(port=443, input_dir: str = None):
    """Run sim rest server for the given nos, version and hostname"""

    api_key = "xxXYHqhFIAWAlWTulizbXtfVfvV5ETfNynHxAlV3ZEUTtrUNKZBDY3aKmCFC"

    auth_response = f"""<response status="success">
    <result>
        <key>{api_key}</key>
    </result>
</response>"""

    username = password = "vagrant"

    routes = web.RouteTableDef()

    @routes.get('/api/')
    async def panos_cmd(request):
        req_type = request.query.get("type", "")
        req_auth_key = request.headers.get("X-PAN-KEY", "") or \
            request.query.get("key", "")

        # authentication with user and pass to get api key
        if req_type == "keygen":
            user = request.query.get("user", "")
            passwd = request.query.get("password", "")
            if user == username and passwd == password:
                return web.Response(text=auth_response)

            raise web.HTTPForbidden()

        # cmd queries
        if req_type == "op" and req_auth_key == api_key:
            xml_cmd = request.query.get("cmd", "")

            if xml_cmd == "":
                raise web.HTTPBadRequest()

            cmd_dict = xmltodict.parse(xml_cmd)
            cmd = get_filename_from_cmd(cmd_dict)
            return web.FileResponse(f"{input_dir}/{cmd}.xml")

    ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    ssl_context.load_cert_chain(
        'suzieq/config/etc/cert.pem', 'suzieq/config/etc/key.pem')

    app = web.Application()
    app.add_routes(routes)
    logging.basicConfig(level=logging.DEBUG)
    web.run_app(app, ssl_context=ssl_context, port=port)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-p", "--listening-port", type=int, default=10000,
        help="Listening port of the ssh server (default: 10000)")
    parser.add_argument(
        "-d", "--input-dir", type=str, default=None,
        help="Input dir to search for host files")
    args = parser.parse_args()

    if not Path(args.input_dir).exists():
        print(f'ERROR: Path {args.input_dir} does not exist, aborting')
        sys.exit(1)

    run_server(port=args.listening_port, input_dir=args.input_dir)



import time

import pandas as pd
from nubia import command

from suzieq.cli.nubia_patch import argument
from suzieq.cli.sqcmds.command import SqTableCommand
from suzieq.sqobjects.tables import TablesObj


@command('table', help='get data about data captured for various tables')
class TableCmd(SqTableCommand):
    """Meta information about the various data gathered"""

    def __init__(
        self,
        engine: str = "",
        hostname: str = "",
        start_time: str = "",
        end_time: str = "",
        view: str = "",
        namespace: str = "",
        format: str = "",  # pylint: disable=redefined-builtin
        columns: str = "default",
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            sqobj=TablesObj,
        )

    @ command("describe")
    @ argument("table", description="interface name to qualify")
    def describe(self, table: str = "tables"):
        """
        Describe fields in table
        """

        if not table:
            df = pd.DataFrame({'error': ['ERROR: Must specify a table']})
            return self._gen_output(df)

        if self.columns != ['default']:
            df = pd.DataFrame(
                {'error': ['ERROR: Cannot specify columns for command']})
            return self._gen_output(df)

        now = time.time()
        df = self.sqobj.describe(table=table)
        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)

        return self._gen_output(df, dont_strip_cols=True)



import pandas as pd

from suzieq.engines.pandas.engineobj import SqPandasEngine


class ArpndObj(SqPandasEngine):
    '''Backend pandas engine class to support manipulating ARP/ND table'''

    @staticmethod
    def table_name():
        '''Table name'''
        return 'arpnd'

    def get(self, **kwargs) -> pd.DataFrame:
        """Retrieve the arpnd table info providing address prefix filtering"""

        prefix = kwargs.pop('prefix', [])
        columns = kwargs.pop('columns', [])
        user_query = kwargs.pop('query_str', '')
        view = kwargs.get('view', self.iobj.view)

        addnl_fields = []
        # Always get the ipAddress if there is a filter prefix
        fields = self.schema.get_display_fields(columns)
        if 'ipAddress' not in fields:
            addnl_fields.append('ipAddress')

        self._add_active_to_fields(view, fields, addnl_fields)
        user_query_cols = self._get_user_query_cols(user_query)
        addnl_fields += [x for x in user_query_cols if x not in addnl_fields]

        df = super().get(addnl_fields=addnl_fields, columns=fields, **kwargs)

        query_str = ''
        filter_prefix = ''

        for p in prefix:
            query_str += (f'{filter_prefix} '
                          f'@self._is_in_subnet(ipAddress,"{p}")')
            filter_prefix = 'or'

        if query_str:
            df = df.query(query_str)

        df = self._handle_user_query_str(df, user_query)

        return df.reset_index(drop=True)[fields]

    def summarize(self, **kwargs):
        """Summarize ARPND info across namespace"""
        self._summarize_on_add_field = [
            ('deviceCnt', 'hostname', 'nunique'),
            ('arpNdEntriesCnt', 'ipAddress', 'count'),
            ('uniqueArpEntriesCnt', 'ipAddress', 'nunique'),
            ('uniqueOifCnt', 'oif', 'nunique')]

        self._summarize_on_add_with_query = [
            ('arpEntriesCnt', '@self._check_ipvers(ipAddress, 4)',
                'ipAddress'),
            ('v6NDEntriesCnt', '@self._check_ipvers(ipAddress, 6)',
                'ipAddress'),
            ('v6NDGlobalEntriesCnt',
                '@self._check_ipvers(ipAddress, 6) and '
                '@self._is_in_subnet(ipAddress, "fe80::/10")', 'ipAddress'),
            ('v6NDLLAEntriesCnt',
                '@self._check_ipvers(ipAddress, 6) and not '
                '@self._is_in_subnet(ipAddress, "fe80::/10")',
                'ipAddress'),
            ('remoteV4EntriesCnt',
                'state == "remote" and @self._check_ipvers(ipAddress, 4)',
                'ipAddress'),
            ('staticV4EntriesCnt', 'state == "permanent" and '
                '@self._check_ipvers(ipAddress, 4)', 'ipAddress'),
            ('remoteV6EntriesCnt',
                'state == "remote" and @self._check_ipvers(ipAddress, 6)',
                'ipAddress'),
            ('staticV6EntriesCnt',
                'state == "permanent" and @self._check_ipvers(ipAddress, 6)',
                'ipAddress'),
            ('failedEntriesCnt', 'state == "failed"', 'ipAddress'),
            ('failedV4EntriesCnt',
                'state == "failed" and @self._check_ipvers(ipAddress, 4)',
                'ipAddress'),
            ('failedV6EntriesCnt',
                'state == "failed" and @self._check_ipvers(ipAddress, 6)',
                'ipAddress')]

        self._check_empty_col = 'arpNdEntriesCnt'
        return super().summarize(**kwargs)



#!/usr/bin/env python

import sys
import pyarrow.parquet as pq
import pyarrow as pa
from suzieq.shared.schema import Schema, SchemaForTable
import logging
import pandas as pd
from pathlib import Path


def convert_dir(input_dir: str, output_dir: str, svcschema: SchemaForTable):
    """Convert the data into a single file and write it out"""

    defaults = {
        pa.string(): "",
        pa.int32(): 0,
        pa.int64(): 0,
        pa.float32(): 0.0,
        pa.float64(): 0.0,
        pa.date64(): 0.0,
        pa.bool_(): False,
        pa.list_(pa.string()): ['-'],
        pa.list_(pa.int64()): [],
    }

    df = pd.read_parquet(input_dir, use_legacy_dataset=True)
    sqschema = svcschema.get_raw_schema()
    arrow_schema = svc_schema.get_arrow_schema()

    for column in filter(lambda x: x['name'] not in df.columns, sqschema):
        df[column['name']] = column.get('default', defaults[column['type']])

    # convert all dtypes to whatever is desired
    for column in df.columns:
        if column in arrow_schema:
            df[column] = df[column].astype(arrow_schema.field(column)
                                           .type.to_pandas_dtype())

    # If there's the original ifname saved up, then eliminate this unnecessary
    # field as this model is no longer necessary

    if 'origIfname' in df.columns:
        if 'ifname' in df.columns:
            df = df.drop(columns=['ifname']) \
                   .rename(columns={'origIfname': 'ifname'})
        elif 'oif' in df.columns:
            df = df.drop(columns=['oif']) \
                   .rename(columns={'origIfname': 'oif'})

    table = pa.Table.from_pandas(df, schema=arrow_schema,
                                 preserve_index=False)
    partition_cols = svcschema.get_partition_columns()

    if 'norifcnReason' in df.columns:
        df.rename({'notifcnReason': 'notificnReason'}, inplace=True)

    pq.write_to_dataset(
        table,
        root_path=output_dir,
        partition_cols=partition_cols,
        version="2.0",
        compression='ZSTD',
        row_group_size=100000,
    )

    logger.info(f'Wrote converted {input_dir}')


if __name__ == "__main__":
    if len(sys.argv) < 4:
        print('Usage: convert_parquet <input dir> <output_dir> <schema_dir>')
        sys.exit(1)

    input_dir = Path(sys.argv[1])
    output_dir = sys.argv[2]
    schemas = Schema(sys.argv[3])
    service = input_dir.parts[-1]
    svc_schema = SchemaForTable(service, schema=schemas)

    logging.basicConfig(stream=sys.stdout, level=logging.WARNING)
    logger = logging.getLogger('sq-converter')
    convert_dir(input_dir, output_dir, svc_schema)



from abc import ABC, abstractmethod
from suzieq.shared.sq_plugin import SqPlugin


class SqEngineObj(SqPlugin, ABC):
    '''Interface class for adding analyzer engine to Suzieq'''

    def __init__(self):
        pass

    @property
    @abstractmethod
    def name(self):
        '''Get the name of the engine obj'''
        raise NotImplementedError

    @abstractmethod
    def get(self, **kwargs):
        '''Retrieve the data given the constraints provided for the table'''
        raise NotImplementedError

    @abstractmethod
    def summarize(self, **kwargs):
        '''Summarize table info'''
        raise NotImplementedError

    @abstractmethod
    def unique(self, **kwargs):
        '''Return unique values or value counts for provided column'''
        raise NotImplementedError

    @abstractmethod
    def aver(self, **kwargs):
        '''Run pre-defined checks on the data'''
        raise NotImplementedError

    @abstractmethod
    def top(self, **kwargs):
        '''Return top values associated with a numeric column'''
        raise NotImplementedError

    @abstractmethod
    def lpm(self, **kwargs):
        '''Run IP LPM algorithm for given address'''
        raise NotImplementedError

    @abstractmethod
    def find(self, **kwargs):
        '''Return endpoint tracker info'''
        raise NotImplementedError




from suzieq.engines.base_engine import SqEngineObj


def get_sqengine(engine_name: str, table_name: str):
    """Return either the top level engine obj or a specific engine obj

    Args:
        engine_name (str): Name of the engine
        table_name (str): Table specific engine object

    Raises:
        Exception: ModuleNotFoundError if the engine is not found

    Returns:
        [class]: SqEngineObj
    """
    engine_obj = SqEngineObj.get_plugins(engine_name).get(engine_name, None)
    if not engine_obj:
        raise ModuleNotFoundError(f"Engine {engine_name} not found")

    if table_name:
        table_obj = engine_obj.get_plugins(table_name).get(table_name, None)
    else:
        table_obj = engine_obj

    if not table_obj:
        raise ModuleNotFoundError(f"Table {table_name} not found for "
                                  f"engine {engine_name}")
    return table_obj


__all__ = ['get_sqengine']



from nubia import command
from suzieq.cli.nubia_patch import argument

from suzieq.cli.sqcmds.command import SqTableCommand
from suzieq.sqobjects.macs import MacsObj


@command("mac", help="Act on MAC Table data")
@argument("vlan", description="VLAN(s). space separated")
@argument("macaddr",
          description="MAC address(es), in quotes, space separated")
@argument("remoteVtepIp",
          description="Remote VTEP IP(s), space separated; use any for all")
@argument("bd", description="Bridge Domain(s), space separated")
@argument("local", type=bool,
          description="filter entries with no remoteVtep")
@argument("moveCount", description="num of times this MAC has moved")
class MacCmd(SqTableCommand):
    """MAC address table information"""

    def __init__(
            self,
            engine: str = "",
            hostname: str = "",
            start_time: str = "",
            end_time: str = "",
            view: str = "",
            namespace: str = "",
            format: str = "",  # pylint: disable=redefined-builtin
            query_str: str = ' ',
            columns: str = "default",
            vlan: str = '',
            macaddr: str = '',
            remoteVtepIp: str = '',
            bd: str = '',
            local: bool = False,
            moveCount: str = ''
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=MacsObj,
        )

        self.lvars = {
            'vlan': vlan.split(),
            'macaddr': macaddr.split(),
            'remoteVtepIp': remoteVtepIp.split(),
            'bd': bd.split(),
            'local': local,
            'moveCount': moveCount
        }



from dataclasses import dataclass

import networkx as nx
import numpy as np
import pandas as pd

from suzieq.engines.pandas.engineobj import SqPandasEngine
from suzieq.shared.utils import build_query_str

# TODO:
# topology for different VRFs?
# iBGP vs eBGP?
# color by device type?
# physical topology without LLDP -- is this possible?
# how to draw multiple topologies
# be able to ask if a node has neighbors by type (physical, overlay, protocol,
# etc) questions
# * without knowing hierarchy, labels or tags it's unclear how to group things
#   for good picture
# how could we add state of connection (like Established) per protocol


class TopologyObj(SqPandasEngine):
    '''Backend class to operate on virtual table, topology, with pandas'''

    @staticmethod
    def table_name():
        '''Table name'''
        return 'topology'

    # pylint: disable=attribute-defined-outside-init
    def _init_dfs(self, namespaces):
        """Initialize the dataframes used for the given namespace"""

        self._if_df = self._get_table_sqobj('interfaces') \
            .get(namespace=namespaces, state="up",
                 columns=['namespace', 'hostname', 'ifname', 'ipAddressList',
                          'ip6AddressList', 'state', 'type', 'master',
                          'macaddr'])

        if self._if_df.empty:
            return

        self._if_df['vrf'] = self._if_df.apply(
            lambda x: x['master'] if x['type'] not in ['bridge', 'bond_slave']
            else 'default', axis=1)
        self._if_df['vrf'] = np.where(self._if_df['vrf'] == '', 'default',
                                      self._if_df['vrf'])
        self._if_df = self._if_df.explode('ipAddressList') \
                                 .explode('ip6AddressList') \
                                 .fillna({'ipAddressList': '0.0.0.0/0',
                                          'ip6AddressList': '::0/0'})
        self._if_df['plen'] = self._if_df.ipAddressList.str.split('/').str[1]
        self._if_df['ipAddress'] = self._if_df.ipAddressList.str.split(
            '/').str[0]
        self._if_df = self._if_df.drop(
            columns=['ipAddressList', 'ip6AddressList'])

        self.nses = self._if_df['namespace'].unique()

    # pylint: disable=too-many-statements
    def get(self, **kwargs):
        '''DIsplay the topology info'''

        @dataclass(frozen=True)
        class Services:
            '''minor class for use below. Could have been a named tuple'''
            name: str
            extra_args: dict
            extra_cols: list
            augment: any

        self._namespaces = kwargs.get("namespace", self.ctxt.namespace)
        hostname = kwargs.pop('hostname', [])
        user_query = kwargs.pop('query_str', '')
        polled = kwargs.pop('polled', '')
        via = kwargs.pop('via', [])
        ifname = kwargs.pop('ifname', '')
        peerHostname = kwargs.pop('peerHostname', [])
        columns = kwargs.pop('columns', ['default'])
        asn = kwargs.pop('asn', '')
        area = kwargs.pop('area', '')
        vrf = kwargs.pop('vrf', '')
        afi_safi = kwargs.pop('afiSafi', '')

        self._init_dfs(self._namespaces)
        if self._if_df.empty:
            return pd.DataFrame({'error': ['No interfaces data found']})
        self.lsdb = pd.DataFrame()
        self._a_df = pd.DataFrame()
        self._ip_table = pd.DataFrame()

        fields = self.schema.get_display_fields(columns)

        if (asn or afi_safi) and area:
            raise AttributeError(
                'Cannot provide asn/afiSafi and area at the same time')

        if any(x in columns for x in ['asn', 'afiSafi']) or asn or afi_safi:
            # When you don't constrain the via to bgp when asn is explicitly
            # requested, we can have some issues. unique or any sort of sorting
            # on the asn column will fail because all rows without BGP == true
            # will have asn as either an undefined value, 0 or blank. The net
            # result of this will be to confuse the user. Think of all the pain
            # automation scripts will have to endure for this.
            if via and via != ['bgp']:
                raise AttributeError(
                    'Cannot provide any via except bgp with asn')
            via = ['bgp']

        if area or 'area' in columns:
            if via and via != ['ospf']:
                raise AttributeError(
                    'Cannot provide any via except ospf with area')
            via = ['ospf']

        if not via:
            via = ['lldp', 'bgp', 'ospf']

        self.services = [
            Services('lldp', {}, ['ifname', 'vrf'],
                     self._augment_lldp_show),
            Services('arpnd', {},
                     ['ipAddress', 'macaddr', 'ifname', 'vrf', 'arpndBidir'],
                     self._augment_arpnd_show),
            Services('bgp', {'state': 'Established', 'asn': asn,
                             'afiSafi': afi_safi, 'vrf': vrf,
                             'columns': ['*']},
                     ['vrf', 'asn', 'peerAsn'],
                     self._augment_bgp_show),
            # Services('evpnVni', {}, 'peerHostname',
            #    'vni', self._augment_evpnvni_show),
            Services('ospf', {'state': 'full', 'area': area, 'vrf': vrf},
                     ['ifname', 'vrf', 'area'], self._augment_ospf_show),
        ]

        if via:
            self.services = [x for x in self.services if x.name in via]
        key = 'peerHostname'

        for srv in self.services:
            if 'columns' not in srv.extra_args:
                srv.extra_args['columns'] = ['default']
            extra_cols = list(srv.extra_cols)
            df = self._get_table_sqobj(srv.name).get(
                **kwargs,
                **srv.extra_args
            )
            if not df.empty:
                if srv.augment:
                    df = srv.augment(df)
                cols = ['namespace', 'hostname', key]
                if extra_cols:
                    cols = cols + extra_cols
                df = df[cols]
                df.insert(len(df.columns), srv.name, True)
                if self.lsdb.empty:
                    self.lsdb = df
                else:
                    self.lsdb = self.lsdb.merge(df, how='outer')
            else:
                self.lsdb[srv.name] = False

        self.lsdb = self.lsdb.fillna('').reset_index(drop=True)
        self._find_polled_neighbors(polled)
        if self.lsdb.empty:
            return self.lsdb

        if 'ipAddress' not in self.lsdb.columns:
            self.lsdb['ipAddress'] = ''

        self.lsdb = self.lsdb[~self.lsdb.hostname.isna()] \
                        .drop_duplicates() \
                        .rename({'ipAddress': 'peerIP', 'macaddr': 'peerMac'},
                                axis=1, errors='ignore') \
                        .dropna(subset=['peerHostname', 'peerIP'], how='all') \
                        .fillna({'peerHostname': 'unknown',
                                 'ifname': 'unknown', 'arpnd': False,
                                 'peerIP': '', 'peerMac': '',
                                 'arpndBidir': False,
                                 'bgp': False, 'ospf': False,
                                 'lldp': False, 'vrf': 'N/A'})

        self.lsdb['vrf'] = np.where(self.lsdb.vrf == "bridge", "-",
                                    self.lsdb.vrf)
        self.lsdb['peerHostname'] = np.where(
            self.lsdb.peerHostname == "unknown", self.lsdb.peerIP,
            self.lsdb.peerHostname)
        self.lsdb = self.lsdb.drop(columns=['peerIP', 'peerMac'],
                                   errors='ignore')
        self.lsdb = self.lsdb.query('peerHostname != ""')

        # Apply the appropriate filters
        if not self.lsdb.empty:
            self.lsdb = self.lsdb.fillna('')
            query_str = build_query_str([], self.schema, ignore_regex=False,
                                        hostname=hostname,
                                        peerHostname=peerHostname,
                                        ifname=ifname, asn=asn, area=area,
                                        vrf=vrf)
            if query_str:
                self.lsdb = self.lsdb.query(query_str)

        if user_query and not self.lsdb.empty:
            self.lsdb = self._handle_user_query_str(self.lsdb, user_query)

        fields = [x for x in self.lsdb.columns if x in fields]
        return self.lsdb[fields].reset_index(drop=True)

    def _find_polled_neighbors(self, polled):
        if self.lsdb.empty:
            return

        hosts = set(self.lsdb.hostname)
        peer_hosts = set(self.lsdb.peerHostname)
        unpolled_neighbors = peer_hosts.difference(hosts)

        self.lsdb['polled'] = True
        self.lsdb.loc[self.lsdb.peerHostname.isin(unpolled_neighbors),
                      'polled'] = False

        if polled and polled.lower() == 'true':
            self.lsdb = self.lsdb.query('polled')
        elif polled.lower() == 'false':
            self.lsdb = self.lsdb.query('~polled')

    def _create_graphs_from_lsdb(self):
        self.graphs = {}
        for ns, df in self.lsdb.groupby(by=['namespace']):
            attrs = [srv.name for srv in self.services
                     if srv.name in df.columns]
            self.graphs[ns] = nx.from_pandas_edgelist(
                df, 'hostname', 'peerHostname', attrs, nx.MultiGraph)

    # TODO: eventually this needs to move to ospf after we figure out the
    #   schema augmentation story
    def _augment_ospf_show(self, df):
        if not df.empty:
            df = df.query('adjState != "passive"').reset_index(drop=True)
            df['area'] = df['area'].astype(str)
        return df

    def _augment_lldp_show(self, df):
        if not df.empty and not self._if_df.empty:
            df = df[df.peerHostname != '']
            df = df.merge(
                self._if_df[['namespace', 'hostname', 'ifname', 'vrf']],
                on=['namespace', 'hostname', 'ifname'], how='outer')

        return df

    def _augment_bgp_show(self, df):
        return df

    def _augment_evpnvni_show(self, df):
        if not df.empty:
            df = df.explode('remoteVtepList').dropna(how='any')
            df = df.rename(columns={'remoteVtepList': 'peerIP'})
            df = df.merge(self.ip_table, on=['namespace', 'peerIP'],
                          how='left').dropna(how='any').drop_duplicates()
        return df

    def _augment_arpnd_show(self, df):

        if not df.empty:
            # weed out entries that are not reachable
            df = df.query('state != "failed"').reset_index(drop=True)
            # Get the VRF
            df = df.merge(
                self._if_df[['namespace', 'hostname', 'ifname', 'master']],
                left_on=['namespace', 'hostname', 'oif'],
                right_on=['namespace', 'hostname', 'ifname'],
                suffixes=['', '_y'], how='outer') \
                .rename(columns={'master': 'vrf'}) \
                .dropna(subset=['ipAddress'])

            df['vrf'] = np.where(df['vrf'] == '', 'default', df['vrf'])
            df = df.drop(columns=['vlan', 'oif', 'mackey', 'remoteVtepIp',
                                  'timestamp_y'], errors='ignore')

            # address are where I find the neighbor, these own the mac
            addr = self.address_df[['namespace', 'hostname', 'macaddr']] \
                .drop_duplicates() \
                .rename(columns={'hostname': 'peerHostname'})
            df = df.merge(addr, on=['namespace', 'macaddr'], how='left') \
                   .dropna(subset=['hostname', 'ipAddress']) \
                   .rename(columns={'oif': 'ifname'}) \
                   .drop_duplicates()

            # Use MAC table entries to find the local port for a MAC on an SVI
            mac_df = self._get_table_sqobj('macs')\
                         .get(namespace=self._namespaces, local=True,
                              columns=['namespace', 'hostname', 'vlan',
                                       'macaddr', 'oif', 'remoteVtepIp'])
            df = df.merge(mac_df,
                          on=['namespace', 'hostname', 'macaddr'],
                          how='outer') \
                .dropna(subset=['hostname'])
            df['ifname'] = np.where(
                df['oif'].isnull(), df['ifname'], df['oif'])

            df['arpndBidir'] = df.apply(
                lambda x, y:
                not y.query(f'namespace=="{x.namespace}" and '
                            f'hostname=="{x.peerHostname}" and '
                            f'peerHostname=="{x.hostname}"').empty,
                args=(df,), axis=1)

        self._arpnd_df = df
        return self._arpnd_df

    @property
    def address_df(self) -> pd.DataFrame():
        '''Return pandas dataframe of the address table'''

        if self._a_df.empty:
            self._a_df = self._if_df
        return self._a_df

    @property
    def ip_table(self):
        '''Return a dataframe of '''
        if self._ip_table.empty:
            addr = self.address_df
            if not addr.empty:
                self._ip_table = addr[['namespace',
                                       'hostname', 'ipAddressList']]
                self._ip_table = self._ip_table.explode(
                    'ipAddressList').dropna(how='any')
                self._ip_table = self._ip_table \
                                     .rename(
                                         columns={'ipAddressList': 'peerIP',
                                                  'hostname': 'peerHostname'})
                self._ip_table['peerIP'] = self._ip_table['peerIP'] \
                                               .str.replace("/.+", "")
                self._ip_table = \
                    self._ip_table[self._ip_table['peerIP'] != '-']
        return self._ip_table

    def summarize(self, **kwargs):
        '''Summarize the topology info'''
        self.get(**kwargs)
        if self.lsdb.empty:
            return self.lsdb

        self.ns = {}
        for i in self.nses:
            self.ns[i] = {}
        self._create_graphs_from_lsdb()
        self._analyze_lsdb_graph()

        return pd.DataFrame(self.ns)

    def _analyze_lsdb_graph(self):
        for ns in self.nses:
            for name in [srv.name for srv in self.services if
                         nx.get_edge_attributes(self.graphs[ns], srv.name)]:

                G = nx.Graph([(s, d, data) for s, d, data in
                              self.graphs[ns].edges(data=True)
                              if data[name] is True])
                if G.nodes:
                    self.ns[ns][f'{name}_number_of_nodes'] = len(G.nodes)
                    self.ns[ns][f'{name}_number_of_edges'] = len(G.edges)
                    if not nx.is_connected(G):
                        self.ns[ns][f'{name}_is_fully_connected'] = False
                        self.ns[ns][f'{name}_center'] = False
                    else:
                        self.ns[ns][f'{name}_is_fully_connected'] = True
                        self.ns[ns][f'{name}_center'] = nx.barycenter(G)

                    self.ns[ns][f'{name}_self_loops'] = list(
                        nx.nodes_with_selfloops(G))

                    self.ns[ns][f'{name}_number_of_disjoint_sets'] = len(
                        list(nx.connected_components(G)))

                    self.ns[ns][f'{name}_degree_histogram'] = \
                        nx.degree_histogram(G)

                    # too many degrees => the column gets too big
                    if len(self.ns[ns][f'{name}_degree_histogram']) > 6:
                        self.ns[ns][f'{name}_degree_histogram'] = '...'

                else:
                    for k in [f'{name}_is_fully_connected', f'{name}_center',
                              f'{name}_self_loops',
                              f'{name}_number_of_disjoint_sets',
                              f'{name}_degree_histogram',
                              f'{name}_number_of_nodes',
                              f'{name}_number_of_edges']:
                        self.ns[ns][k] = None



import asyncio
from asyncio.subprocess import Process


async def monitor_process(process: Process, label: str):
    """Wait and monitor the process given as input printing the stdout and
    stderr content

    Args:
        process (Process): the process to wait and monitor
        label (str): the label identifying the process
    """
    stdout_read = asyncio.create_task(process.stdout.readline())
    stderr_read = asyncio.create_task(process.stderr.readline())
    tasks = [stdout_read, stderr_read]
    while tasks:
        done, pending = await asyncio.wait(
            tasks,
            return_when=asyncio.FIRST_COMPLETED
        )
        tasks = list(pending)
        for d in done:
            if d.result():
                print(f'[{label}]: {d.result().decode()[:-1]}')
                if d == stdout_read:
                    stdout_read = asyncio.create_task(
                        process.stdout.readline())
                    tasks.append(stdout_read)
                else:
                    stderr_read = asyncio.create_task(
                        process.stderr.readline())
                    tasks.append(stderr_read)



import pandas as pd
from pandas.core.dtypes.dtypes import DatetimeTZDtype

from suzieq.sqobjects.basicobj import SqObject
from suzieq.shared.utils import humanize_timestamp


class DeviceObj(SqObject):
    '''The object providing access to the device table'''

    def __init__(self, **kwargs):
        super().__init__(table='device', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'columns', 'os',
                                'vendor', 'model', 'status', 'version',
                                'query_str', 'ignore_neverpoll']
        self._valid_arg_vals = {
            'status': ['alive', 'dead', 'neverpoll',
                       '!alive', '!dead', '!neverpoll']
        }
        self._unique_def_column = ['model']

    def humanize_fields(self, df: pd.DataFrame, _=None) -> pd.DataFrame:
        '''Humanize the timestamp and boot time fields'''
        if df.empty:
            return df

        # Convert the bootup timestamp into a time delta
        if 'bootupTimestamp' in df.columns:
            if not isinstance(df.bootupTimestamp.dtype, DatetimeTZDtype):
                df['bootupTimestamp'] = humanize_timestamp(
                    df['bootupTimestamp']*1000,
                    self.cfg.get('analyzer', {}).get('timezone', None))

        return super().humanize_fields(df)
